/*****************************************************************
KERNEL PORTION OF WINDOWS-BASED TEXT EDITING PROGRAM
   to accompany article "TEXT EDITOR ARCHITECTURES AND ALGORITHMS"
   by Ray Valdes in the April 1993 issue of Dr Dobbs Journal.
------------------------------------------------------------------
   Copyright 1993 by Ray Valdes.
   Limited license granted for non-commercial and educational purposes.
   All other rights reserved.
------------------------------------------------------------------

   This is *not* the complete working version, which is currently still
   under (re)-construction. I thought I would have the program complete
   by the time the article was printed, but it has since grown to 
   *double* the original size (from 4000 to 8000 lines of code
   for the total program, while the kernel has grown from 2000 to
   about 4700 lines of code).

   I've been adding various optimizations, MDI support, a icon toolbar, 
   updates for Windows 3.1, and also trying to make the coding style 
   more consistent (following my idiosyncratic coding convention).
   However, a lot of what Microsoft calls the "Windows grunge" is not yet
   working.  Even so, few of these changes are relevant to the subject 
   of my article, which was on general algorithms and architectures for
   text editors. In my article, I promised to provide "a concrete
   illustration" of the discussion on architectures and algorithms.

   So the following 4700 lines of code provide an example how to implement
   many key routines within an industrial-strength text editor engine.
   Unfortunately, the code is not standalone, and will produce compilation
   errors in its present form.
   
   So it's best if you view this code as an_extremely_detailed_pseudocode_
   representation_of_a_text_editor, rather than as an actual implementation.

   I'll make a working implementation available at a future date, as
   time permits.  Sorry for the change, but the code here is much
   greater amount of text editor source than originally promised.
   (If you do want proven editor code, this not hard to come by.  There
   are a number of complete, tested, working implementations of text
   editors available on CompuServe and on the Internet. On CompuServe,
   the MSWIN32 forum has source code MicroEmacs for Windows/NT.
   The IBMPRO forum has code for DOS MicroEmacs, FreeMacs, and
   a bunch of other editors.  For a small, clean, fast implementation
   in C, check out F.Eng's CHI editor in Borlands C/C++ DOS Forum.
   Austin CodeWorks in Austin TX sells, on floppy disk, the source
   code to 20 different public-domain or publicly available editors,
   such as MicroEmacs, Jove, Elvis, Grief, etc.)
   
   My code illustrates the following aspects of implementing a
   text editor:

   * maintaining a document/view architecture (although this implementation
     only allows one view per document, it can be extended to multiple
     views per document without too much additional work).

   * how to maintain two parallel streams, one of text characters and
     the other of text attributes.  The attribute stream is a sparse
     representation using an array of attribute records. An attribute
     record is allocated and used to point into the text stream whenever
     there is any change in font/size/face combination.
     
   * how to maintain a simple text-to-screen map of line breaks. This
     is roughly similar to the attribute stream in that a separate
     array points into the stream of text characters.

   * providing a small public interface (API) to clients of this engine.
     All public routines are prefixed with "ed_", as in 
     ed_CreateNewDocument() or ed_UpdateView().  The rest of the 
     implementation is encapsulated in private (i.e., static) routines
     that are prefixed with "priv_" as in priv_ReformatEntireDocument().
     
   * isolating the graphics primitives from the engine (all environment-
     dependent routines are prefixed with "gp_" as in "gp_TextOut()").
     I've tried to minimize dependencies on Windows-specific constructs
     such as HWNDs and memory handles.  The memory subsystem, in
     particular, is encapsulated by an interface layer whose functions
     are prefixed with "mem_" as in mem_AllocHandle().  It should
     be portable to other platforms such as Macintosh and Unix/Motif
     without too much pain.  There is still some Windows-specific
     code, which you can recognize by #ifdef WINDOWS_SPECIFIC_CODE,
     as well as by use of functions prefixed with "win_", such as
     win_GetDC(), which are basically macro substitions for the
     equivalent functions in the native Windows API.

   * how to process keystrokes which are both commands as well as text.
     As an illustration, there is a very small subset of the Emacs command
     set implemented, see the priv_DoEmacsCommand() function.
     
   * how to handle mouse-down, mouse-move and mouse-up events and highlight
     a region of text on the screen. See ed_OnMouseDownMsg() and
     related functions.
     
   * scrolling the view. See the function ed_ScrollView().
   
   * maintaining a mapping between text stream positions and screen
     (x,y) locations. See priv_MapTextToScreen().

   * various straightforward strategies for incremental reformatting
     and redisplay.

   There are some important items missing from my example.  The most
   glaring omission is any kind of sophisticated buffer management.

   Regarding buffer management, in my article I discuss 3 strategies
   for managing the text stream:
       
   1. Dumb and simple, using memmov() to shift bytes in memory
      one keystroke at a time.

   2. Better, but limited in storage capacity, using a buffer-gap
      approach on a single large block of RAM to minimize per-keystroke
      processing.

   3. More sophisticated but requiring more implementation effort,
      using a "virtual memory" scheme implemented in software, to
      eliminate RAM-size constraints.

   In my article, I present a buffer-gap module I wrote that is 
   derived from code by Joe Allen.  That code is included in this
   ARC file, but has not been integrated with this text editing
   engine.  This engine uses the dumb approach in item #1 above.
   
   The design goals of this engine are not for handling large
   amounts of text, but for illustrating mouse-event handling
   and multi-font processing, and also as a laboratory for
   testing various optimization strategies for incremental
   reformatting and redisplay.
   
   There are 3 optimization strategies being tested here,
   which are enabled by compile-time switches (preprocessor
   #defines), as follows:
   
   1. CHAR_WIDTH_CACHE -- code that is delimited by this #ifdef
      directive implements a cache for character widths, on a
      per font/size/face combination.  This helps avoid having
      to make calls to the Windows API.
      
   2. XCOORD_CACHE -- this conditional code implements a cache
      for horizontal positions (x-coordinates) on a given line.
      So when a mouse-down event occurs, we can determine very
      quickly which character corresponds to the mouse down. 
      This structure is also useful in centering lines of text,
      or in setting text flush right.
   
   3. OUTPUT_BITMAP_CACHE -- this optimization may actually be
      of questionable value on most machines, but seemed at one
      time to be necessary for slow machines running Windows (e.g.,
      286). This code maintains a cached image of the current 
      line of text, including various font and pointsize changes.
      So when the user types a keystroke, the program uses this
      cached bitmap to minimize rendering the rest of the line,
      and instead uses BitBlt to output to screen.
   

Note: Tab setting for the following listing is 8 spaces per tab stop.
*****************************************************************/

#include "w_config.h"
#include <windows.h> 
#include "std_defs.h"
#include "g_mem.h"
#include "g_debug.h"

/****************************************************************
  ED_CONFIG.H -- Compile-Time Options for text editing module
*****************************************************************/

/* ---- these compile-time switches enable feature-set options-----*/
#undef  MULTI_FONT		/* enable multiple fonts/sizes/styles */
#undef  EMACS_COMMANDS		/* enable handling of subset of Emacs cmds */

/* ---- these compile-time switches enable various optimizations-----*/
#undef  XCOORD_CACHE		/* enable caching of xcoordinates */
#undef  CHAR_WIDTH_CACHE	/* enable caching of character widths */
#undef  OUTPUT_BITMAP_CACHE	/* enable output bitmap image of curr line */


/****************************************************************
STD_DEFS.H --- some #defines for my peculiar coding convention
****************************************************************/

#define BLOCK
#define PUBLIC
#define PRIVATE static
#define OPTIM	register

/****************************************************************
ED_STRUCT.H
****************************************************************/

typedef int           TEXTPOS;
typedef char           TEXTCH;
typedef TEXTCH FAR * LPTEXTCH;
typedef int           TEXTPOS;

typedef int ERR_STAT;  /* error/status return value */

typedef enum 
{
    FLUSH_RIGHT, FLUSH_LEFT, FLUSH_CENTER 
} FLUSH;

/******************* The Document Structure *********************/

/* to have multiple views onto the same document, the
 * following structure should be broken up into substructures
 * with appropriate pointers/handles.  Right now there is
 * only a one-to-one correspondence between View & Document
 */

typedef	struct
{
	/*------The View-----*/
	HWND    hWnd;
	RECT	viewRect;
	int	active;
	int	lineHeight;
	int	fontAscent;

	/*-----Selection range-----*/
	RECT	selRect;
	POINT	selPoint;
	int	selStart;
	int	selEnd;

	/*-----Attributes----------*/
	int	attrMode;		// per paragraph
	FLUSH	attrFlush;		// per paragraph
	int	attrTypefont;		// per text run
	char	attrTypeface;		// per text run
	int	attrPointsize;		// per text run

#ifdef MULTI_FONT
	HMEMBLK  hFormat;
#endif

	/*-----The Data Model----------*/
	RECT	galleyRect;		/* in document space */
	int	numTextChars;
	int	numPhysLines;
	int	numLines;  /*logical lines*/
	HMEMBLK	hLineMap;
	HMEMBLK	hTxt;

} DOCSTRUC, 
  FAR * LPDOC; 

/****************************************************************/

typedef struct 
{
	int	lineMap[1];
} LINEMAP,
  LINEMAP FAR * LPLINEMAP;

typedef HMEMBLK HDOC;

typedef int XCOORD;
typedef XCOORD FAR * LPXCOORD;

/****************************************************************/

#define    INVALID_LINE		(-1)
#define    INVALID_COORD	(-32767)
#define    BACKSPACE_CH		(8)
#define    PARA_BREAK_CH	(0xB6)
#define    LINES_QUANTUM	(4)
#define    MIN_HTXT_SIZE	(16)
#define    MIN_LINEMAP_SIZE	(16)
#define    MIN_SCRAP_SIZE	(128)
#define    SIZE_OF_TEXTCHBUF	(1024)
#define    DEFAULT_LINE_HEIGHT	(14)
#define    DEFAULT_CHAR_WIDTH	(8)

typedef enum
{
    LS_DELETE_SEGMENT,
    LS_INSERT_SEGMENT
} LINEOP; 				/* for lineMap[] calculations */

#define    LS_INVALID_POSITION  (-1)	/* for lineMap[] calculations */

/****************************************************************/


		/*-------Public Interface----------------------------*/
PUBLIC VOID	ed_InitModule		(void);
PUBLIC VOID	ed_ReformatDoc		(HDOC);
PUBLIC HDOC	ed_CreateNewDocument	(LPRECT, LPRECT,HWND);
PUBLIC VOID	ed_DeleteDocument	(HDOC);
PUBLIC VOID	ed_UpdateView		(LPRECT, HDOC);
PUBLIC VOID	ed_InsertString		(LPTEXTCH, long, HDOC);
PUBLIC VOID	ed_OnMouseDownMsg	(LPPOINT,BOOL,HDOC);
PUBLIC VOID	ed_OnDoubleClickMsg	(HDOC,int,int);

PUBLIC VOID	ed_ScrollView		(HDOC,int,int);

PUBLIC ERR_STAT ed_FromWinClipboard	(void);
PUBLIC ERR_STAT ed_ToWinClipboard	(void);

PUBLIC VOID	ed_PasteFromCutBuff	(HDOC hDoc);
PUBLIC VOID	ed_CopySelTextToCutBuff	(HDOC hDoc);
PUBLIC VOID	ed_DeleteSelText	(HDOC hDoc);

#ifdef MULTI_FONT
PUBLIC VOID	ed_ChangeFont		(HDOC,int);
PUBLIC VOID	ed_ChangeFace		(HDOC,int);
PUBLIC VOID	ed_ChangeSize		(HDOC,int);
#endif

		/*-------Internal Modules (Basic Stuff)-------------*/
PRIVATE VOID	priv_UpdateView		(LPRECT,LPDOC);
PRIVATE VOID	priv_SetTextData	(LPTEXTCH,long,LPDOC);
PRIVATE HMEMBLK priv_GetTextData	(LPDOC);
PRIVATE VOID	priv_InsertString	(LPTEXTCH, long, LPDOC);
PRIVATE VOID	priv_SetSelectionRange	(long,long,LPDOC);
PRIVATE VOID	priv_Activate		(LPDOC);
PRIVATE VOID	priv_Deactivate		(LPDOC);
PRIVATE VOID	priv_SetFlushAttr	(int,LPDOC);
PRIVATE VOID	priv_ScrollView		(LPDOC,int,int);

PRIVATE VOID	priv_ReformatEntireDocument	(LPDOC);

		/*------- Screen To Text Mapping---------------------*/
PRIVATE VOID	priv_MapScreenToText	(LPDOC, LPPOINT,LPINT, LPPOINT);
PRIVATE BOOL	priv_MapTextToScreen	(LPDOC, int, LPPOINT);
PRIVATE int	priv_LocateLine		(LPDOC, int);
PRIVATE VOID	priv_UnshowSelection	(LPDOC);
PRIVATE VOID	priv_ShowSelection	(LPDOC);

		/*--------- Caret Manager --------------------------*/
PRIVATE VOID	priv_InitCaret		(void);
PRIVATE VOID	priv_DestroyCaret	(void);
PRIVATE VOID	priv_DrawCaret		(LPPOINT, LPDOC);

		/*-----------Reformatting of Text----------------*/
PRIVATE VOID	priv_IncrementalReformat(LPDOC,int);
PRIVATE VOID	priv_DeleteOneChar	(LPDOC, int);

PRIVATE VOID	priv_InsertOneChar	(LPDOC,TEXTCH,int);
PRIVATE int	priv_UpdateLineStarts	(LPDOC,int);
PRIVATE VOID	priv_ReformatForOne	(LPDOC,HMEMBLK,int,int);
PRIVATE VOID	priv_FormatUntilSynch	(LPDOC,int,int,int);
PRIVATE VOID	priv_MarkLineStarts	(int,LPDOC,int,int);
PRIVATE VOID	priv_ResetPrevLine	(void);



PRIVATE HMEMBLK priv_GetCutBuffHandle	(void);
PRIVATE long	priv_GetCutBuffLen	(void);
PRIVATE VOID	priv_SetCutBuffLen	(MEMSIZE);

PRIVATE VOID	priv_DeleteSelText	(LPDOC);
PRIVATE VOID	priv_CutSelTextToCutBuff(LPDOC);
PRIVATE VOID	priv_CopySelTextToCutBuff(LPDOC);
PRIVATE VOID	priv_PasteFromCutBuff	(LPDOC);

PRIVATE VOID	ft_DrawFmtLine		(LPDOC,HMEMBLK,int,int);
PRIVATE VOID	ft_PrintFmtList		(LPDOC);
PRIVATE VOID	ft_DeleteFromFormat	(LPDOC,int,int);
PRIVATE VOID	ft_AddToFormat		(LPDOC,int,int);

PRIVATE  VOID  ft_NewFmtBlock		(LPDOC);
PRIVATE  VOID  ft_InsertFmtRecord	(HMEMBLK,LPDOC,int,int,int);
PRIVATE  VOID  ft_ChangeParam		(LPDOC,int,int);
PRIVATE  VOID  misc_ShiftNBytes		(MEMADDR,MEMADDR,int);

PRIVATE  LPFORMAT  ft_GetFirstFState 	(LPFORMAT,int,LPBOOL );
PRIVATE  LPFORMAT  ft_GetNextFState 	(LPFORMAT,int,LPBOOL );


		/*------- EMACS Commands-------------*/
PRIVATE VOID	priv_DoEmacsCommand	(TEXTCH, LPDOC);

		/*------- Debugging-------------------*/
PRIVATE VOID	priv_ShowStructures	(LPDOC);

		/*------- Xcoord Cache---------------*/
PRIVATE VOID	priv_InitXCoordCache		(void);
PRIVATE VOID	priv_ResetXCoordCache		(void);
PRIVATE VOID	priv_InvalidateCachedLineXCoord	(LPDOC);
#ifdef XCOORD_CACHE
PRIVATE HMEMBLK priv_GetLineHandleXCoord	(LPDOC, LPTEXTCH,int);
#endif

		/*------- Char Width Cache-----------*/
PRIVATE VOID	priv_InitCharWidthCache		(void);
PRIVATE VOID	priv_RenewCharWidthCache		(int,int,int);
PRIVATE int	priv_CalcLinebreak		(LPTEXTCH,LPTEXTCH,int);

		/*------- Highlighting --------------*/
PRIVATE VOID	priv_DoHighlightRange		(LPDOC,LPPOINT,LPPOINT);

				/* macro alias for above function */ 
#define    priv_DoUnhighlightRange priv_DoHighlightRange 

PRIVATE BOOL	ed_IsAnyDocActive	(void);


/***************Graphics Primitives Layer (gp_) *****************/
/* 
 * the following functions are not part of the Text Editor Kernel
 * but are called by it.  They provide a information-hiding layer
 * over the Windows GDI (or equivalent, such as Macintosh QuickDraw)
 */

PRIVATE VOID	gp_InvertRect		(LPDOC,int,int,int,int);
PRIVATE VOID	gp_ClearRect		(LPDOC,int,int,int,int);
PRIVATE VOID	gp_TextOutAt		(int,int,LPDOC,HMEMBLK,int,int);
PRIVATE BOOL	kb_IsThereKeystroke	(LPDOC);
PRIVATE VOID	gp_ScrollWindow		(LPDOC, int,int,LPRECT);
PRIVATE VOID	gp_SetClipRect		(LPDOC,LPRECT);
PRIVATE VOID	gp_RestoreClipRect	(LPDOC);
PRIVATE VOID	gp_DrawLineOfText	(LPDOC, LPTEXTCH, int, int, BOOL);

PRIVATE VOID    gp_ValidateScreenRect	(LPRECT);
PRIVATE VOID	gp_SetRect		();
PRIVATE VOID	gp_SectRect		();
PRIVATE VOID	gp_EraseRect		();
PRIVATE VOID	gp_OffsetRect		();
PRIVATE VOID	gp_GetPort		();
PRIVATE VOID	gp_CharWidth		();
/*
 * the following are functions that are very close to
 * the Windows API level of abstraction.  That is,
 * they are not particularly platform independent.
 * The hope is to someday remove these from the code
 * and replace them with higher-level functions.
 */

#ifdef WINDOWS_SPECIFIC_CODE
PRIVATE HDC	win_GetDC		();
PRIVATE VOID	win_ReleaseDC		();
PRIVATE VOID 	win_StretchBlt		();
PRIVATE HANDLE	win_SelectObject	();
PRIVATE HANDLE	win_DeleteObject	();
PRIVATE HBITMAP win_CreateBitmap	();
PRIVATE VOID	win_OpenClipboard	();
PRIVATE VOID	win_CloseClipboard	();
PRIVATE VOID	win_GetClipboardData	();
PRIVATE VOID	win_ScrollWindow	();
#endif

/****************************************************************
ED_GLOBAL.C
****************************************************************/

PRIVATE BOOL isDoubleDown;

/*-------------------------------------------------*/
PUBLIC BOOL ed_InitModule(VOID)
/*-------------------------------------------------*/
{
    priv_InitCaret();
    priv_InitXCoordCache();
#ifdef OUTPUT_BITMAP_CACHE
    priv_InitBitmapCache();
#endif
    return priv_InitCutBuff();
}


/*---------------------------------------------------------*/
PUBLIC VOID ed_ReformatDoc(HDOC hDoc)
/*---------------------------------------------------------*/
{
   LPDOC lpDoc;
   lpDoc = (LPDOC) mem_GraspHandle( hDoc);  
   priv_ReformatEntireDocument(lpDoc);
   mem_UngraspHandle( hDoc);     
}

/*---------------------------------------------------------*/
PUBLIC VOID ed_UpdateView(
    LPRECT arg_rect,
    HDOC   hDoc)
/*---------------------------------------------------------*/
{
    LPDOC lpDoc;
    lpDoc = (LPDOC) mem_GraspHandle( hDoc); 
    priv_UpdateView(arg_rect,lpDoc);
    mem_UngraspHandle(hDoc);
}

/*---------------------------------------------------------*/
PUBLIC HDOC    ed_CreateNewDocument(
    LPRECT     galley_rect,
    LPRECT     view_rect,
    HWND       hWnd)
/*---------------------------------------------------------*/
{
    LPDOC lpDoc;
    HDOC    hDoc;
    HMEMBLK hTxt;
    HMEMBLK hLineMap;
    
    hDoc = (HDOC) mem_AllocHandle(sizeof(DOCSTRUC));
    hTxt     = mem_AllocHandle(MIN_HTXT_SIZE);
    hLineMap = mem_AllocHandle(sizeof(TEXTPOS)*LINES_QUANTUM);

    if(!hDoc || !hTxt || !hLineMap)
	err_FatalError("ed_CreateNewDocument: hDoc or hTxt is NULL!");
    
    lpDoc = (LPDOC) mem_GraspHandle(hDoc);


	/*------The View-----*/
    lpDoc->hWnd = hWnd; 
    lpDoc->viewRect	= *view_rect;	/*structure copy*/
    gp_SetRect(&(lpDoc->selRect),0,0,0,0);
    lpDoc->lineHeight	= (int)	DEFAULT_LINE_HEIGHT;
    lpDoc->fontAscent	= 0;

	/*-----Selection range-----*/
    lpDoc->selPoint	= *((LPPOINT)galley_rect);
    lpDoc->selStart	= 0;
    lpDoc->selEnd	= 0;
    lpDoc->active	= FALSE;

	/*-----Attributes----------*/
    lpDoc->attrFlush	= FLUSH_LEFT;
    lpDoc->attrTypefont	= 0;
    lpDoc->attrTypeface	= 0;
    lpDoc->attrMode	= 0;
    lpDoc->attrPointsize= 0;

A

#ifdef MULTI_FONT
    lpDoc->hFormat	=   0;
#endif

	/*-----The Data Model----------*/
    lpDoc->galleyRect	= *galley_rect;	/*structure copy*/
    lpDoc->numTextChars	= 0;
    lpDoc->numPhysLines = LINES_QUANTUM; /*internal use: # of physlines*/
    lpDoc->numLines     = 0; /* number of logical lines */
    lpDoc->hTxt	        =  hTxt;
    lpDoc->hLineMap     = hLineMap;
    mem_GraspHandle(hLineMap)->lineMap[0]=0;	
    mem_UngraspHandle(hLineMap);
    
    mem_UngraspHandle(hDoc);
    return hDoc;
}

/****************************************************************/
PUBLIC VOID ed_DeleteDocument(HDOC hDoc)
{
    LPDOC lpDoc;
    lpDoc = (LPDOC) mem_GraspHandle(hDoc);
    if (lpDoc->hTxt     ) mem_DeleteHandle(lpDoc->hTxt);
    if (lpDoc->hLineMap ) mem_DeleteHandle(lpDoc->hLineMap);    
#ifdef MULTI_FONT
    if (lpDoc->hFormat) mem_DeleteHandle(lpDoc->hFormat); 
#endif
    mem_DeleteHandle( hDoc);
}


/*---------------------------------------------------------*/
PUBLIC VOID   ed_OnCharMsg(
     TEXTCH   ch,
     HDOC     hDoc)
/*---------------------------------------------------------*/
{
    OPTIM LPDOC	lpDoc;
    int		insert_length, new_length, sel_length;
    POINT	bottom_pt;
    BOOL	dont_set_clip;


#ifdef EMACS_COMMANDS    
    if(ch < 32 && ch != BACKSPACE_CH && ch != NEWLINE_CH && ch != CR)
     {
	priv_DoEmacsCommand(ch,hDoc); 
	return; 
    }
#endif    
    
    lpDoc = (LPDOC) mem_GraspHandle(hDoc);   
    priv_ShowStructures(lpDoc);

    sel_length = (lpDoc->selEnd - lpDoc->selStart);

    priv_UnshowSelection(lpDoc);

    /* get bottommost pt, before bashing old numTextChars */
    priv_MapTextToScreen(lpDoc, lpDoc->numTextChars, &bottom_pt);

    /* for now, ANY content change invalidates the line in lineoffset cache. 
     * Future optimizations can change this logic.
     */
    priv_InvalidateCachedLineXCoord(lpDoc);

    /* check for special cases: backspace-delete, simple single char insert*/
    if (ch == BACKSPACE_CH)
    {
	if (sel_length > 0)   /* case: need to delete a segment of text */
	{
	    priv_DeleteSelText(lpDoc); /* calls ReformatEntireDocument() */
	    priv_UpdateView(&(lpDoc->viewRect),lpDoc);	    
	    mem_UngraspHandle( hDoc);
	    return;
	}
	 /* case: need to delete just one character, and NOT at loc 0 */
	else if(lpDoc->selEnd > 0)
	{
	    priv_DeleteOneChar(lpDoc,bottom_pt.y);
	}
	priv_ShowSelection(lpDoc);
	mem_UngraspHandle( hDoc);
	return;
    }
    else if(sel_length==0)	/* case of simply inserting one char */
    {
	dont_set_clip = (  lpDoc->galleyRect.left  >= lpDoc->viewRect.left
			&& lpDoc->galleyRect.right <= lpDoc->viewRect.right);
	if(!dont_set_clip) gp_SetClipRect(hDoc,&(lpDoc->viewRect));

	priv_InsertOneChar(lpDoc,ch,bottom_pt.y);

	priv_ShowSelection(lpDoc);

	if(!dont_set_clip) gp_RestoreClipRect(hDoc);
	mem_UngraspHandle( hDoc);
	return;
     }
    else     /* case of inserting one char, but there is some text selected */
    {
	insert_length = 1; 
    }

#ifdef MULTI_FONT
    if(lpDoc->hFormat)
    {
	ft_DeleteFromFormat(lpDoc,lpDoc->selStart,lpDoc->selEnd);
	ft_AddToFormat(     lpDoc,lpDoc->selStart,insert_length);
    }
#endif

    /* case of inserting a character with some text selected */    
    if (insert_length <= sel_length)
    {
 	HMEMBLK hTxt = lpDoc->hTxt;

	/* sub case: text to be inserted <= selected text */
	
	new_length = lpDoc->selStart + insert_length + 
			    (lpDoc->numTextChars - lpDoc->selEnd);

	priv_MarkLineStarts(LS_DELETE_SEGMENT,
			    lpDoc,lpDoc->selStart,lpDoc->selEnd  );
	priv_MarkLineStarts(LS_INSERT_SEGMENT,
			    lpDoc,lpDoc->selEnd  ,lpDoc->selEnd+1);
	
	/* call mem_ChangeHandleSize() AFTER the bytemoves */

	BLOCK
	{
	    LPTEXTCH lpTxt = (LPTEXTCH) mem_GraspHandle(hTxt);
 	    OPTIM LPTEXTCH src = lpTxt + lpDoc->selEnd;
	    OPTIM LPTEXTCH dst = lpTxt + lpDoc->selStart 
						    + insert_length;   
	    OPTIM int   cnt = lpDoc->numTextChars - lpDoc->selEnd;
	
	    /* don't move bytes if insert_length==sel_length */
	    if(dst < src)
		while(cnt--) *dst++ = *src++;
	
	    dst  = lpTxt + lpDoc->selStart;
	    *dst = ch;
	    mem_UngraspHandle( hTxt);
	}
	hTxt = lpDoc->hTxt = mem_ChangeHandleSize(hTxt, 
		    (MEMSIZE) (MAX(MIN_HTXT_SIZE,new_length)));
    }
    else  /*----- subcase: insert_length > sel_length && sel_length > 0 ---*/
    {
	HMEMBLK hTxt = lpDoc->hTxt;


	new_length = lpDoc->numTextChars + (insert_length - sel_length);

	/* do mem_ChangeHandleSize before copy, so we have room to copy into */
	hTxt = lpDoc->hTxt = mem_ChangeHandleSize(lpDoc->hTxt, 
				(MEMSIZE) (MAX(MIN_HTXT_SIZE,new_length)));
	priv_MarkLineStarts(LS_DELETE_SEGMENT,
			    lpDoc,lpDoc->selStart,lpDoc->selEnd  );
	priv_MarkLineStarts(LS_INSERT_SEGMENT,
			    lpDoc,lpDoc->selEnd  ,lpDoc->selEnd+1);
	BLOCK
	{
	    LPTEXTCH lpTxt = (LPTEXTCH) mem_GraspHandle(hTxt);
	    OPTIM LPTEXTCH src = lpTxt + lpDoc->numTextChars - 1;
	    OPTIM LPTEXTCH dst = lpTxt + new_length - 1;
	    OPTIM int      cnt = lpDoc->numTextChars - lpDoc->selEnd;
	    while(cnt--) *dst-- = *src--;
	
	    dst  = lpTxt + lpDoc->selStart;
	    *dst = ch;  /* put the char in the hole */
	    mem_UngraspHandle( hTxt);
	}
    }
    lpDoc->numTextChars = new_length;	
    lpDoc->selEnd      -= sel_length;
    lpDoc->selEnd      += insert_length;
    lpDoc->selStart     = lpDoc->selEnd;

    dont_set_clip = (    lpDoc->galleyRect.left  >= lpDoc->viewRect.left
		      && lpDoc->galleyRect.right <= lpDoc->viewRect.right);
    if(!dont_set_clip) gp_SetClipRect(hDoc,&(lpDoc->viewRect));

	    /* reformat and redraw the text */
    priv_IncrementalReformat(lpDoc,bottom_pt.y); 
    priv_ShowSelection(lpDoc);
    mem_UngraspHandle( hDoc);

    if(!dont_set_clip) gp_RestoreClipRect(hDoc);
    
    return;
}

/*---------------------------------------------------------*/
PRIVATE VOID priv_ShowStructures( OPTIM LPDOC lpDoc)
/*---------------------------------------------------------*/
{
    int i,y;

    for ( i = 0, y = lpDoc->galleyRect.top
	; i <= lpDoc->numLines
	; i++, y += lpDoc->lineHeight)
    {
	/* show structures*/;
    }
}


/****************************************************************/
PUBLIC VOID ed_SetTextData(
    LPTEXTCH src,
    long segment_length,
    HDOC hDoc)
{
    OPTIM LPDOC 	lpDoc;
    HMEMBLK 		hTxt;
    LPTEXTCH		lpTxt;

    lpDoc = (LPDOC) mem_GraspHandle(hDoc);
#ifdef MULTI_FONT
    if(lpDoc->hFormat)
    {
	Warning("SetTextData: hFormat already exists!");
	mem_DeleteHandle( (lpDoc->hFormat));
	lpDoc->hFormat = 0;
    }
#endif

    if(lpDoc->hTxt)
    {
	Warning("SetTextData: hTxt already exists!");
	mem_DeleteHandle(lpDoc->hTxt);
    }

    lpDoc->hTxt = mem_AllocHandle(segment_length);

    hTxt = lpDoc->hTxt;
    lpTxt = (LPTEXTCH) mem_GraspHandle(hTxt);

    BLOCK
    {
	OPTIM LPTEXTCH dest  = lpTxt;
	OPTIM int count = (int) segment_length;
	while (count--) *dest++ = *src++;
    }
    mem_UngraspHandle( hTxt);

			    /* in theory, we lose bits here */
    lpDoc->numTextChars = (int) segment_length; 
    priv_ReformatEntireDocument(lpDoc); /* must now reformat text */
    mem_UngraspHandle(hDoc);
}

/****************************************************************/
PUBLIC HMEMBLK ed_GetTextData(HDOC hDoc)
{
    LPDOC lpDoc;
    HMEMBLK hTxt;
    lpDoc = (LPDOC) mem_GraspHandle(hDoc);
    hTxt = lpDoc->hTxt;
    mem_UngraspHandle(hDoc);
    return hTxt;
}


/****************************************************************/

PRIVATE HDOC active_hDoc;

PRIVATE int ed_IsAnyDocActive(void)
{
    return (active_hDoc != NULL);
}

/****************************************************************/
PRIVATE VOID priv_SetSelectionRange(
    long start, 
    long end,
    HDOC hDoc)
{
    OPTIM LPDOC lpDoc;

    lpDoc = (LPDOC) mem_GraspHandle( hDoc);   

    if(lpDoc->active)    priv_UnshowSelection(lpDoc);    

    lpDoc->selStart = start;
    lpDoc->selEnd   = (end > lpDoc->numTextChars) ? lpDoc->numTextChars : end;

    if(lpDoc->active)    priv_ShowSelection(lpDoc);    

    mem_UngraspHandle( hDoc);
}

/****************************************************************/
PRIVATE VOID priv_Activate(HDOC hDoc)
{
    if (hDoc==NULL) err_FatalError("Activate: hDoc is NULL!");
    if (lpDoc->active || active_hDoc != NULL)
	Warning("Activate called twice!");

    priv_ResetXCoordCache();
    priv_RenewCharWidthCache(
	lpDoc->attrTypefont,
	lpDoc->attrTypeface,
	lpDoc->attrPointsize);

#ifdef OUTPUT_BITMAP_CACHE /*10/7 -- bitmaps */
    priv_ResetBitmapCache();
#endif
    active_hDoc	     = hDoc;
    lpDoc->active   = TRUE;
}

/****************************************************************/
PRIVATE VOID priv_Deactivate(HDOC hDoc)
{
    if (!lpDoc->active || active_hDoc==NULL)
	Warning("Deactivate: already inactive!");

    active_hDoc	     = NULL;
    lpDoc->active   = FALSE;

    priv_DestroyCaret();
}

/****************************************************************/
PRIVATE VOID priv_SetFlushAttr(
    int attrFlush, 
    HDOC hDoc)
{
    lpDoc->attrFlush = attrFlush;
    priv_ReformatEntireDocument(hDoc);    /* need to reformat text now */
}


/****************************************************************/
PRIVATE VOID priv_UpdateView(
    LPRECT arg_rect,
    LPDOC  lpDoc)
{
    int		start_pos,count;
    int		x,y;
    int		cur_line;
    BOOL	dont_set_clip;

    RECT	vis_rect,tmp_rect,galley_rect;

     /* redraw any text that falls within the specified rectangle */
    if (arg_rect->right <= arg_rect->left) 
    {
	Warning("UpdateView: width < 0!");
	arg_rect->right = arg_rect->left + (lpDoc->lineHeight * 2);
    }

    priv_ShowStructures(lpDoc);

    priv_ResetXCoordCache();
    priv_RenewCharWidthCache(
	lpDoc->attrTypefont,
	lpDoc->attrTypeface,
	lpDoc->attrPointsize);

    priv_DestroyCaret();

    galley_rect = lpDoc->galleyRect; /*structure copy */
    galley_rect.bottom = MIN(lpDoc->viewRect.bottom, arg_rect->bottom);

    if(    !gp_SectRect (&(lpDoc->viewRect),arg_rect,&tmp_rect)
	|| !gp_SectRect (&galley_rect,&tmp_rect,&vis_rect)
	)
    {
	return;
    }
    
    dont_set_clip = (    lpDoc->galleyRect.left  >= lpDoc->viewRect.left
                      && lpDoc->galleyRect.right <= lpDoc->viewRect.right);
    if(!dont_set_clip) gp_SetClipRect(hDoc,&(lpDoc->viewRect));

    lpTxt = (LPTEXTCH) mem_GraspHandle(lpDoc->hTxt);

    for (    y = galley_rect.top,    cur_line=0
    	  ;  y < vis_rect.bottom
	  ;  y += lpDoc->lineHeight, cur_line++
	  )
    {
	if(    y >= vis_rect.top
	    && y + lpDoc->lineHeight <= vis_rect.bottom )
	{
	    if(lpDoc->active) /* clear rect only if textnode is active */
	    {
		gp_ClearRect(lpDoc, 
		       vis_rect.left,  y,
		       vis_rect.right, lpDoc->lineHeight);
	    }
	    if (cur_line < lpDoc->numLines)
	    {
		gp_DrawLineOfText(lpDoc,lpDoc->hTxt,cur_line,y,FALSE);
	    }
	}
    } 
    mem_UngraspHandle( (lpDoc->hTxt));
    if(lpDoc->active) priv_ShowSelection(lpDoc);
    mem_UngraspHandle( hDoc);
    if(!dont_set_clip) gp_RestoreClipRect(hDoc);
    return;
}

/****************************************************************/
PUBLIC VOID ed_ScrollView(
    HDOC hDoc,
    int dh,
    int dv)
{
    LPDOC lpDoc;
    LPDOC lpDoc;
    lpDoc = (LPDOC) mem_GraspHandle( hDoc);  
    priv_ScrollView(lpDoc,dh,dv);
    mem_UngraspHandle(hDoc);
}

/****************************************************************/
PRIVATE VOID priv_ScrollView(
    LPDOC lpDoc,
    int dh,
    int dv)
{
    RECT scroll_rect, horz_erase_rect, vert_erase_rect;
    int dh_abs, dv_abs;

    if(dh==0 && dv==0) return;

	/* structure copies */
    scroll_rect = horz_erase_rect = vert_erase_rect = lpDoc->viewRect;
   
    priv_UnshowSelection(lpDoc);
    dv_abs = ABS(dv);
    dh_abs = ABS(dh);    
	
    if(dv != 0) /* case vert scroll */
    {
    	if(dv < 0) /* case vert scroll UP */
    	{
    		scroll_rect.top = 
		    MIN(scroll_rect.bottom, scroll_rect.top + dv_abs);
    		vert_erase_rect.top =
		    MAX(lpDoc->viewRect.top, lpDoc->viewRect.bottom - dv_abs);
    		vert_erase_rect.top   -= 
			((vert_erase_rect.top - lpDoc->galleyRect.top) 
			% lpDoc->lineHeight);
    		/*scroll_rect.bottom = vert_erase_rect.top;  */
    	}
    	else if(dv != 0) /* case vert scroll DOWN */
    	{
    		scroll_rect.bottom = 
		    MAX(scroll_rect.top, scroll_rect.bottom - dv_abs);
    		if(scroll_rect.bottom > lpDoc->galleyRect.top)
    		{
		    scroll_rect.bottom -= 
    			((scroll_rect.bottom - lpDoc->galleyRect.top) 
    			% lpDoc->lineHeight);
    		}
    		vert_erase_rect.bottom = 
		    MIN(lpDoc->viewRect.bottom, lpDoc->viewRect.top + dv_abs);
    	 }
    }
    if (dh != 0) /* case: horz scroll */
    {
	if(dh < 0) /* case horz scroll LEFT */
       	{
		scroll_rect.left = 
		    MIN(scroll_rect.right, scroll_rect.left - dh);
    		horz_erase_rect.left = 
		    MAX(lpDoc->viewRect.left, lpDoc->viewRect.right - dh_abs);
	}
	else /* case horz scroll RIGHT */
	{
		scroll_rect.right =
		    MAX(scroll_rect.left,scroll_rect.right - dh);
		horz_erase_rect.right = 
		    MIN(lpDoc->viewRect.right,lpDoc->viewRect.left + dh_abs);
	}
    }
    gp_ScrollWindow( lpDoc->hWnd, dh,dv, &scroll_rect);

    if(dh && horz_erase_rect.right > horz_erase_rect.left) 
		EraseRect(&horz_erase_rect);
    if(dv && vert_erase_rect.bottom > vert_erase_rect.top) 
		EraseRect(&vert_erase_rect);

	/* Now do the update for the exposed piece and prevent a Show */
    gp_OffsetRect( &(lpDoc->galleyRect), dh, dv);

    if(dv) priv_UpdateView(&vert_erase_rect,lpDoc); 
    if(dh) priv_UpdateView(&horz_erase_rect,lpDoc);

    gp_ValidateScreenRect(&(lpDoc->viewRect));

    return;
}

/****************************************************************
	SHOWING AND UNSHOWING SELECTED TEXT
****************************************************************/

/*--------------------------------------------------------*/
PRIVATE VOID  priv_UnshowSelection(OPTIM LPDOC lpDoc)   
/*--------------------------------------------------------*/
{
    POINT   caret_pt;
    POINT   other_pt;
    BOOL    end_within_view;
    BOOL    start_within_view;

    priv_DestroyCaret();
    
    if(lpDoc->selStart < lpDoc->selEnd) 
    {
	end_within_view = priv_MapTextToScreen(lpDoc,lpDoc->selEnd,&caret_pt);

	start_within_view = 
		    priv_MapTextToScreen(lpDoc, lpDoc->selStart,&other_pt);

	if(!start_within_view && !end_within_view)
	{
	    return;
	}
	priv_DoHighlightRange(lpDoc,other_pt,caret_pt);	
    }
    return;
}


/*--------------------------------------------------------*/
PRIVATE VOID  priv_ShowSelection(OPTIM LPDOC lpDoc)
/*--------------------------------------------------------*/
{
    POINT caret_pt;
    POINT other_pt;
    BOOL end_within_view,start_within_view;
    
    end_within_view = priv_MapTextToScreen(lpDoc, lpDoc->selEnd,&caret_pt);

    if(lpDoc->selStart==lpDoc->selEnd && end_within_view)
    {
	priv_DrawCaret(caret_pt,lpDoc);
    }
    else
    {
	start_within_view = 
		    priv_MapTextToScreen(lpDoc, lpDoc->selStart,&other_pt);
	if( !start_within_view && !end_within_view)
	{
	    return;
	}
	priv_DoHighlightRange(lpDoc,other_pt,caret_pt);	
    }
}


/****************************************************************
ED_CUT.C -- Management of CutBuff (Internal Clipboard)
****************************************************************/


typedef struct
{
    MEMSIZE    cb_physize;
    MEMSIZE    cb_length;
    HMEMBLK    cb_handle;
} TXTCUTBUFF;

PRIVATE TXTCUTBUFF theCutBuff;

/*-------------------------------------------------*/
PRIVATE BOOL priv_InitCutBuff(VOID)
/*-------------------------------------------------*/
{
    theCutBuff.cb_physize = MIN_SCRAP_SIZE;
    theCutBuff.cb_handle  = mem_AllocHandle(theCutBuff.cb_physize);
    theCutBuff.cb_length  = 0;
    return (theCutBuff.cb_handle != 0);
}


/*-------------------------------------------------*/
PUBLIC ERR_STAT ed_FromWinClipboard(HWND hAppWnd)
/*-------------------------------------------------*/
{
    HANDLE hMem;
    OPTIM LPTEXTCH src,p;
    OPTIM int count;

    if	(!win_OpenClipboard(hAppWnd))
    {
	return (ERR_STAT) !NULL;
    }
    if (!(hMem = win_GetClipboardData(CF_TEXT)))
    {
	win_CloseClipboard();
	return (ERR_STAT) !NULL;
    }
    src = (LPTEXTCH ) mem_GraspHandle(hMem);
    for(p = src, count = 0; *p; p++, count++)
		/*continue*/;
    theCutBuff.cb_physize =  MAX(count,MIN_SCRAP_SIZE);
    theCutBuff.cb_handle = 
	mem_ChangeHandleSize(theCutBuff.cb_handle,theCutBuff.cb_physize);
    theCutBuff.cb_length = count;

    BLOCK 
    {
	OPTIM LPTEXTCH dest = (LPTEXTCH)mem_GraspHandle(theCutBuff.cb_handle);
	while(count--) *dest++ = *src++;
    }

    mem_UngraspHandle(theCutBuff.cb_handle);
    mem_UngraspHandle(hMem);
}


/*-------------------------------------------------*/
PUBLIC ERR_STAT ed_ToWinClipboard(HWND theAppWindow) 
/*-------------------------------------------------*/
{
    OPTIM  LPTEXTCH src,dest;
    OPTIM count;
    HMEMBLK hMem;

    
    src = (LPTEXTCH ) mem_GraspHandle(theCutBuff.cb_handle);
    hMem = mem_AllocHandle(theCutBuff.cb_length + 1);
    if	(!win_OpenClipboard(theAppWindow))
    {
	return FAILURE;
    }
    if(hMem==NULL)
    {
	win_CloseClipboard();
	mem_UngraspHandle(theCutBuff.cb_handle);
	return FAILURE;
    }
    dest = (LPTEXTCH )mem_GraspHandle(hMem);
    count = theCutBuff.cb_length;
    while(count--) *dest++ = *src++;
    *dest = NULL;
    mem_UngraspHandle(hMem);
    gp_EmptyClipboard();
    gp_SetClipboardData(CF_TEXT, hMem);
    win_CloseClipboard();
    mem_UngraspHandle(theCutBuff.cb_handle);
    return SUCCESS;
}


/*-------------------------------------------------*/
PRIVATE HMEMBLK priv_GetCutBuffHandle(VOID)
/*-------------------------------------------------*/
{
    return theCutBuff.cb_handle;
}


/*-------------------------------------------------*/
PRIVATE long priv_GetCutBuffLen(VOID)
/*-------------------------------------------------*/
{
    return (long)theCutBuff.cb_length;
}

/*-------------------------------------------------*/
PRIVATE VOID priv_SetCutBuffLen(Size length)
/*-------------------------------------------------*/
{
    theCutBuff.cb_length  = 0;
    theCutBuff.cb_physize = MAX(length,MIN_SCRAP_SIZE);
    theCutBuff.cb_handle = mem_ChangeHandleSize(theCutBuff.cb_handle,
	(MEMSIZE) (theCutBuff.cb_physize));
}


/*-------------------------------------------------*/
PRIVATE VOID priv_CutSelTextToCutBuff(HDOC hDoc)
/*-------------------------------------------------*/
{
    OPTIM LPDOC lpDoc;
    int sel_length;
    HMEMBLK hSource,hDest;

    lpDoc = (LPDOC) mem_GraspHandle( hDoc); 
#ifdef MULTI_FONT
    if(lpDoc->hFormat)
	ft_DeleteFromFormat(lpDoc,lpDoc->selStart,lpDoc->selEnd);

#endif

    sel_length = lpDoc->selEnd - lpDoc->selStart;

    theCutBuff.cb_length  = sel_length;

    if(sel_length==0)     /* zero the scrap */
    {
	theCutBuff.cb_physize = MIN_SCRAP_SIZE;
	theCutBuff.cb_handle  = mem_ChangeHandleSize(
	    theCutBuff.cb_handle, 
	    theCutBuff.cb_physize);
	mem_UngraspHandle( hDoc);
	return;
    }
    else if (sel_length > theCutBuff.cb_physize) /* make scrap bigger */
    {
	theCutBuff.cb_physize = MAX(sel_length,MIN_SCRAP_SIZE);
	theCutBuff.cb_handle  = mem_ChangeHandleSize(
	    theCutBuff.cb_handle,
	    theCutBuff.cb_physize);
    }
    hSource = lpDoc->hTxt;	       
    lpSrc = (LPTEXTCH) mem_GraspHandle(hSource);    
    hDest   = theCutBuff.cb_handle;    
    lpDst = (LPTEXTCH) mem_GraspHandle(hDest);

    priv_MarkLineStarts(LS_DELETE_SEGMENT,lpDoc,lpDoc->selStart,lpDoc->selEnd);

	/* transfer over to scrap */
    BLOCK
    {
	OPTIM LPTEXTCH src = lpSrc + lpDoc->selStart;
	OPTIM LPTEXTCH dst = lpDst;
	OPTIM int   cnt = sel_length;

	while(cnt--) *dst++ = *src++;
    
	/* delete from text block */
	src = lpSrc + lpDoc->selEnd ;
	dst = lpSrc + lpDoc->selStart;
	cnt = lpDoc->numTextChars - lpDoc->selEnd;
	while(cnt--) *dst++ = *src++;
    }
    mem_UngraspHandle(hDest);
    mem_UngraspHandle(hSource);        

    lpDoc->numTextChars -= sel_length;
    lpDoc->selEnd   -= sel_length;
    lpDoc->selStart  = lpDoc->selEnd;
    
    lpDoc->hTxt = hSource = mem_ChangeHandleSize(hSource, 
	(MEMSIZE) (MAX(MIN_HTXT_SIZE,lpDoc->numTextChars)));

    mem_UngraspHandle( hDoc);
    priv_ReformatEntireDocument(hDoc);    /* need to reformat text now */

    if(lpDoc->active)
    {
	RECT r;
	r = lpDoc->viewRect; /* structure copy */
	priv_UpdateView(&r,hDoc);	    	
    }
}

/*-------------------------------------------------*/
PUBLIC VOID ed_CopySelTextToCutBuff(HDOC hDoc)
/*-------------------------------------------------*/
{
    LPDOC lpDoc;
    lpDoc = (LPDOC) mem_GraspHandle( hDoc); 
    priv_CopySelTextToCutBuff(lpDoc);
    mem_UngraspHandle(hDoc);
}

/*-------------------------------------------------*/
PRIVATE VOID priv_CopySelTextToCutBuff(LPDOC lpDoc)
/*-------------------------------------------------*/
{
    int sel_length;
    HMEMBLK hSource,hDest;

    sel_length = lpDoc->selEnd - lpDoc->selStart;
    theCutBuff.cb_length  = sel_length;

    if(sel_length==0)     /* zero the scrap */
    {
	theCutBuff.cb_physize = MIN_SCRAP_SIZE;
	theCutBuff.cb_handle   = mem_ChangeHandleSize(
	    theCutBuff.cb_handle,
	    theCutBuff.cb_physize);
	return;
    }
    else if (sel_length > theCutBuff.cb_physize)
    {
	theCutBuff.cb_physize = MAX(sel_length,MIN_SCRAP_SIZE);
	theCutBuff.cb_handle = mem_ChangeHandleSize(
	    theCutBuff.cb_handle,
	    theCutBuff.cb_physize);
    }

    hSource = lpDoc->hTxt;	       
    lpSrc = (LPTEXTCH) mem_GraspHandle(hSource);    
    hDest   = theCutBuff.cb_handle; 
    lpDst = (LPTEXTCH) mem_GraspHandle(hDest);

    BLOCK
    {
	OPTIM LPTEXTCH src = lpSrc + lpDoc->selStart;
	OPTIM LPTEXTCH dst = lpDst;
	OPTIM int   cnt = sel_length;
	while(cnt--) *dst++ = *src++;
    }
    
    mem_UngraspHandle(hDest);
    mem_UngraspHandle(hSource);        
    priv_UpdateView(&(lpDoc->viewRect),hDoc);	    	
    return;
}

/*-------------------------------------------------*/
PUBLIC VOID ed_PasteFromCutBuff(HDOC hDoc)
/*-------------------------------------------------*/
{
    LPDOC lpDoc;
    lpDoc = (LPDOC) mem_GraspHandle( hDoc);  
    priv_PasteFromCutBuff(lpDoc);
    mem_UngraspHandle(hDoc);
}

/*-------------------------------------------------*/
PRIVATE VOID priv_PasteFromCutBuff(LPDOC lpDoc)
/*-------------------------------------------------*/
{
    HMEMBLK hTxt;
    int insert_length, new_length, sel_length;

    insert_length = (int) theCutBuff.cb_length; /*in theory, lose bits here*/
    sel_length    = (lpDoc->selEnd - lpDoc->selStart);
    
#ifdef MULTI_FONT
    if(lpDoc->hFormat)
    {
	if(sel_length) 
	    ft_DeleteFromFormat(lpDoc,lpDoc->selStart,lpDoc->selEnd);
	ft_AddToFormat(         lpDoc,lpDoc->selStart,insert_length);
    }
#endif
	
    /* case: sel_length > 0 and insert_length <= sel_length */
    if (sel_length > 0 && insert_length <= sel_length)
    {

	new_length = lpDoc->selStart + insert_length + 
			    (lpDoc->numTextChars - lpDoc->selEnd);

	/* mem_ChangeHandleSize() is called AFTER byte moves */

	hTxt = lpDoc->hTxt;
	lpTxt = (LPTEXTCH) mem_GraspHandle( hTxt);

	if (sel_length > 0)
	    priv_MarkLineStarts(LS_DELETE_SEGMENT,lpDoc,
				lpDoc->selStart, lpDoc->selEnd);

	priv_MarkLineStarts(LS_INSERT_SEGMENT,lpDoc,
				lpDoc->selEnd, lpDoc->selEnd + insert_length);
	
	BLOCK
	{
	    OPTIM LPTEXTCH src = lpTxt + lpDoc->selEnd;
	    OPTIM LPTEXTCH dst = lpTxt + lpDoc->selStart 
							+ insert_length;
	    OPTIM int   cnt = lpDoc->numTextChars - lpDoc->selEnd;
	
	    if(dst < src) /* don't move bytes if insert_length==sel_length */
		  while(cnt--) *dst++ = *src++;
	
	     src = (LPTEXTCH) mem_GraspHandle(theCutBuff.cb_handle);
	     dst = lpTxt + lpDoc->selStart;
	     cnt = insert_length;
	     while(cnt--) *dst++ = *src++;

	     mem_UngraspHandle(theCutBuff.cb_handle);
	}
	mem_UngraspHandle( hTxt);

	lpDoc->hTxt = hTxt = mem_ChangeHandleSize(hTxt, 
	    (MEMSIZE) (MAX(MIN_HTXT_SIZE,new_length)));
    }
    else    
    {
	if (sel_length==0
	    || (sel_length > 0 && insert_length > sel_length) )
	{
	    new_length = lpDoc->numTextChars + (insert_length - sel_length);
	}
	else
	    err_FatalError("PasteFromCutBuff: invalid case!");

	lpDoc->hTxt = hTxt = mem_ChangeHandleSize(lpDoc->hTxt, 
				(MEMSIZE) (MAX(MIN_HTXT_SIZE,new_length)));

	lpText = (LPTEXTCH) mem_GraspHandle( hTxt);

	if (sel_length > 0)
	    priv_MarkLineStarts(LS_DELETE_SEGMENT,lpDoc,
					lpDoc->selStart, lpDoc->selEnd);

	priv_MarkLineStarts(LS_INSERT_SEGMENT,lpDoc,
			    lpDoc->selEnd, lpDoc->selEnd + insert_length);
			
	BLOCK
	{
	    OPTIM LPTEXTCH src = lpTxt + lpDoc->numTextChars - 1;
	    OPTIM LPTEXTCH dst = lpTxt + new_length - 1;	
	    OPTIM int   cnt = lpDoc->numTextChars - lpDoc->selEnd;
	    
	    while(cnt--) *dst-- = *src--;
	
	    src = (LPTEXTCH) mem_GraspHandle(theCutBuff.cb_handle);
	    dst = lpTxt + lpDoc->selStart;
	    cnt = insert_length;
	    while(cnt--) *dst++ = *src++;

	    mem_UngraspHandle(theCutBuff.cb_handle);	
	}
	mem_UngraspHandle( hTxt);
    }

    lpDoc->numTextChars = new_length;
    lpDoc->selEnd   -= sel_length;
    lpDoc->selEnd   += insert_length;
    lpDoc->selStart  = lpDoc->selEnd;
    mem_UngraspHandle( hDoc);

    priv_ReformatEntireDocument(hDoc);    /* need to reformat text now */

    if(lpDoc->active)
    {
	RECT r;
	r = lpDoc->viewRect; /* structure copy */
	priv_UpdateView(&r,hDoc);
    }
}

/*-------------------------------------------------*/
PUBLIC VOID ed_DeleteSelText(HDOC hDoc)
/*-------------------------------------------------*/
{
    LPDOC lpDoc;
    lpDoc = (LPDOC) mem_GraspHandle( hDoc);  
    priv_DeleteSelText(lpDoc);
    mem_UngraspHandle(hDoc);
}

/*-------------------------------------------------*/
PRIVATE VOID priv_DeleteSelText(LPDOC lpDoc)
/*-------------------------------------------------*/
{
    /* Delete selected text. Unlike Cut operation, do not copy to cutbuff */
    int sel_length;
    
#ifdef MULTI_FONT
    if(lpDoc->hFormat)
	ft_DeleteFromFormat(lpDoc,lpDoc->selStart,lpDoc->selEnd);
#endif
    sel_length = (lpDoc->selEnd - lpDoc->selStart);

    if(sel_length==0) 
    {
	return; 
    }
    else
    {
	HMEMBLK hTxt;
	LPTEXTCH lpTxt;
	OPTIM LPTEXTCH src;
	OPTIM LPTEXTCH dst;
	OPTIM int  cnt;
	
	hTxt = lpDoc->hTxt;	
	lpTxt = (LPTEXTCH) mem_GraspHandle( hTxt);
    
	/* delete from text block */
	src = lpTxt + lpDoc->selEnd;
	dst = lpTxt + lpDoc->selStart;
	cnt = (lpDoc->numTextChars - lpDoc->selEnd);

	while(cnt--) *dst++ = *src++;

	priv_MarkLineStarts(LS_DELETE_SEGMENT,
		    lpDoc,lpDoc->selStart, lpDoc->selEnd);
	mem_UngraspHandle( hTxt);

	lpDoc->hTxt = hTxt = mem_ChangeHandleSize(hTxt, 
		(MEMSIZE) (MAX(MIN_HTXT_SIZE,lpDoc->numTextChars)));

	lpDoc->numTextChars -= sel_length;
	lpDoc->selEnd       -= sel_length;
	lpDoc->selStart      = lpDoc->selEnd;
	priv_ReformatEntireDocument(lpDoc);    /* need to reformat text now */
    }
    return;
}

/*-------------------------------------------------*/
PUBLIC VOID ed_InsertString(
    LPTEXTCH new_text, 
    long segment_length,
    HDOC hDoc)
/*-------------------------------------------------*/
{
    LPDOC lpDoc;
    lpDoc = (LPDOC) mem_GraspHandle( hDoc);    
    priv_InsertString(new_text,segment_length,lpDoc);
    mem_UngraspHandle(hDoc);
}

/*-------------------------------------------------*/
PRIVATE VOID priv_InsertString(
    LPTEXTCH new_text, 
    long     segment_length,
    LPDOC    lpDoc)
/*-------------------------------------------------*/
{
    HMEMBLK hTxt;
    int insert_length, new_length, sel_length;

    insert_length = (int) segment_length; /* in theory, we lose bits here */
    sel_length    = (lpDoc->selEnd - lpDoc->selStart);

#ifdef MULTI_FONT
    if(lpDoc->hFormat)
    {
	if(sel_length) 
	    ft_DeleteFromFormat(lpDoc,lpDoc->selStart,lpDoc->selEnd);
	ft_AddToFormat(lpDoc,lpDoc->selStart,insert_length);
    }
#endif
    
    /* case: sel_length > 0 and insert_length <= sel_length */
    if (sel_length > 0 && insert_length <= sel_length)
    {
	new_length = lpDoc->selStart + insert_length + 
			    (lpDoc->numTextChars - lpDoc->selEnd);

	/* mem_ChangeHandleSize() is called AFTER byte moves */

	hTxt = lpDoc->hTxt;
	lpTxt = (LPTEXTCH) mem_GraspHandle( hTxt);

	BLOCK
	{
	    OPTIM LPTEXTCH src = lpTxt + lpDoc->selEnd;
	    OPTIM LPTEXTCH dst = lpTxt + lpDoc->selStart 
							+ insert_length;
	    OPTIM int   cnt = lpDoc->numTextChars - lpDoc->selEnd;
	
	    if(dst < src)/* don't move bytes if insert_length==sel_length */
		while(cnt--) *dst++ = *src++;
	
	    src = new_text;
	    dst = lpTxt + lpDoc->selStart;
	    cnt = insert_length;
	    while(cnt--) *dst++ = *src++;
	}
	mem_UngraspHandle( hTxt);

	lpDoc->hTxt = hTxt = mem_ChangeHandleSize(hTxt,
	    (MEMSIZE) (MAX(MIN_HTXT_SIZE,new_length)));
	lpDoc->numTextChars = new_length;	
    }
    else    
    {
	if (sel_length==0
	    || (sel_length > 0 && insert_length > sel_length) )
	{
	    new_length = lpDoc->numTextChars + (insert_length - sel_length);
	}
	else
	    err_FatalError("InsertString: invalid case!");

	lpDoc->hTxt = mem_ChangeHandleSize(lpDoc->hTxt, 
			    (MEMSIZE) (MAX(MIN_HTXT_SIZE,new_length)));

	hTxt = lpDoc->hTxt;
	lpTxt = mem_GraspHandle( hTxt);

	BLOCK
	{
	    OPTIM LPTEXTCH src = lpTxt + lpDoc->numTextChars - 1;
	    OPTIM LPTEXTCH dst = lpTxt +    new_length - 1;
	    OPTIM int   cnt = lpDoc->numTextChars - lpDoc->selEnd;
	    
	    while(cnt--) *dst-- = *src--;
	
	    src = new_text;
	    dst = lpTxt + lpDoc->selStart;
	    cnt = insert_length;
	    while(cnt--) *dst++ = *src++;
	}
	mem_UngraspHandle( hTxt);

	lpDoc->numTextChars = new_length;
    }

    priv_ReformatEntireDocument(lpDoc);    /* need to reformat text now */

    if(lpDoc->active)
    {
	priv_UpdateView(&(lpDoc->viewRect),lpDoc);	    	
    }
}

/****************************************************************
    ED_CLICK.C
*****************************************************************/


PRIVATE VOID  priv_OnMouseDownMsg		(LPPOINT,BOOL,LPDOC);
PRIVATE VOID  priv_OnDoubleClickMsg		(LPDOC,int,int);

PRIVATE VOID  set_hi_far_pt			(LPPOINT);
PRIVATE VOID  priv_DoHighlightIncrementally	(LPDOC,LPPOINT);
PRIVATE  VOID  priv_DoHighlightRange		(LPDOC,LPPOINT,LPPOINT);

/* in tbtext.h*/
/*#define priv_UnhighlightRange priv_DoHighlightRange (alias for above func)*/

/*---------------------------------------------------------*/
PUBLIC VOID ed_OnMouseDownMsg(
    LPPOINT		start_pt,
    BOOL 		extend,
    HDOC 		hDoc
)
/*---------------------------------------------------------*/
{
    LPDOC		lpDoc;
    lpDoc = (LPDOC) mem_GraspHandle( hDoc); 
    priv_OnMouseDownMsg(start_pt,extend,lpDoc);
    mem_UngraspHandle(hDoc);
}

/*---------------------------------------------------------*/
PRIVATE VOID priv_OnMouseDownMsg(
    LPPOINT		start_pt,
    BOOL 		extend,
    HDOC 		hDoc
)
/*---------------------------------------------------------*/
{
    POINT		cursor_pt,snapped_cursor,snapped_start;
    int			cur_pos,start_pos,prev_pos;
    BOOL		do_show_selection;

    extern BOOL isDoubleDown;

    if (isDoubleDown)
    {
	isDoubleDown = FALSE;
	priv_OnDoubleClickMsg(lpDoc,start_pt->x,start_pt->y);
	return;
    }
    priv_ShowStructures(lpDoc);
    cursor_pt   = *start_pt;

    priv_UnshowSelection(lpDoc);

    if(extend) /* shift-click*/
    {
	int tmp_pos;
	priv_MapScreenToText(lpDoc,cursor_pt, &tmp_pos, &snapped_cursor);
	if(tmp_pos < lpDoc->selStart)      
		lpDoc->selStart = tmp_pos;
	else
		lpDoc->selEnd   = tmp_pos;

	priv_MapTextToScreen(lpDoc, lpDoc->selStart, &snapped_cursor);
	set_hi_far_pt(snapped_cursor);
	snapped_start = snapped_cursor; /*structure copy */
	start_pos     = tmp_pos;
	prev_pos      = tmp_pos;
	cur_pos       = -1;
	priv_MapTextToScreen(lpDoc, lpDoc->selEnd, &snapped_cursor);	
	priv_DoHighlightIncrementally(lpDoc,snapped_cursor);
    }
    else
    {
	priv_MapScreenToText(lpDoc,cursor_pt, &start_pos, &snapped_cursor);

	set_hi_far_pt(snapped_cursor);	

	snapped_start = snapped_cursor; /*structure copy */
	prev_pos      = start_pos;
	cur_pos       = -1;
	
	/* kick off the highlighting with the current point */
	priv_DoHighlightIncrementally(lpDoc,snapped_cursor);
    }
    
#ifdef WINDOWS_SPECIFIC_CODE
BLOCK  
{
    MSG   msg;

    SetCapture(lpDoc->hWnd);

    do_show_selection = FALSE;
    
    /* Eat mouse messages until a WM_LBUTTONUP is encountered. Meanwhile
     * continue to highlight.
     */
    for (;;)
    {
	win_WaitMessage();
	if (win_PeekMessage(&msg,NULL,WM_MOUSEFIRST,WM_MOUSELAST,PM_REMOVE))
	{
            cursor_pt.x  = LOWORD(msg.lParam);
            cursor_pt.y  = HIWORD(msg.lParam);

            if (msg.message == WM_LBUTTONUP)
	    {
		if(cur_pos== -1) 
		{
		    cur_pos = start_pos;
		    do_show_selection = TRUE;
		}
		break;
	    } /*if*/
	    priv_MapScreenToText(lpDoc,cursor_pt, &cur_pos, &snapped_cursor);
	    if(cur_pos != prev_pos)
	    {
		priv_DoHighlightIncrementally(lpDoc,snapped_cursor);
		prev_pos = cur_pos;
	    }
        } /*if*/
    }/*for*/
    win_ReleaseCapture();
}/*block*/
#endif

#ifdef MACINTOSH_SPECIFIC_CODE
BLOCK
{
    EventRecord		event;
    int			eventmask = (mUpMask);

    do_show_selection = FALSE;

    for(;;)    /*track mouse */
    {
	GetNextEvent(eventmask, &event);
	GetMouse(&cursor_pt);

	if (event.what == mouseUp) 
	{ 
	    if(cur_pos== -1) 
	    {
		cur_pos = start_pos;
		do_show_selection = TRUE;
	    }
	    break;
	}
	priv_MapScreenToText(lpDoc,cursor_pt, &cur_pos, &snapped_cursor);
	if(cur_pos != prev_pos)
	{
	    priv_DoHighlightIncrementally(lpDoc,snapped_cursor);
	    prev_pos = cur_pos;
	}
    }
}
#endif

    if(extend)
    {
	if( cur_pos > lpDoc->selStart)
	{
	    lpDoc->selEnd = cur_pos;
	}
	else if (cur_pos < lpDoc->selStart)
	{
	    lpDoc->selEnd   = lpDoc->selStart;
	    lpDoc->selStart = cur_pos;
	}
    }
    else
    {
	if(start_pos < cur_pos)	
	{
	    lpDoc->selEnd   = cur_pos;
	    lpDoc->selStart = start_pos;
	}
	else
	{
	    lpDoc->selEnd   = start_pos;
	    lpDoc->selStart = cur_pos;
	}
    }
    if( do_show_selection 
	|| lpDoc->selStart == lpDoc->selEnd) 
    {
	priv_ShowSelection(lpDoc);
    }
    return;	
}


/*---------------------------------------------------------*/
PUBLIC VOID  ed_OnDoubleClickMsg(
    HDOC	hDoc,
    int		x,
    int		y)
/*---------------------------------------------------------*/
{
    LPDOC lpDoc;
    lpDoc = (LPDOC) mem_GraspHandle( hDoc);    
    priv_OnDoubleClickMsg(hDoc,x,y);
    mem_UngraspHandle(hDoc);
}

/*---------------------------------------------------------*/
PRIVATE VOID  priv_OnDoubleClickMsg(
    LPDOC	lpDoc,
    int		x,
    int		y)
/*---------------------------------------------------------*/
{
#define is_word_delimiter(c)	((c)==SPACE_CH || (c)==CR || (c)==TAB_CH)

    int		cursor_pos;
    POINT	cursor_pt,start_hilite,end_hilite;
    HMEMBLK	hTxt;
    LPTEXTCH 	start_ptr;
    LPTEXTCH 	end_ptr;
    LPTEXTCH 	cur_ptr;
    OPTIM LPTEXTCH 	text_ptr;
    
    priv_ShowStructures(lpDoc);
    
    priv_UnshowSelection(lpDoc);

    cursor_pt.x = x; 
    cursor_pt.y = y;

    priv_MapScreenToText(lpDoc,cursor_pt, &cursor_pos, &start_hilite);

    hTxt = lpDoc->hTxt; 
    lpTxt = (LPTEXTCH) mem_GraspHandle( hTxt);

    start_ptr = lpTxt;
    text_ptr  = cur_ptr = lpTxt + cursor_pos;
    end_ptr   = lpTxt + lpDoc->numTextChars;

    while(!is_word_delimiter(*text_ptr) && text_ptr > start_ptr)
	text_ptr--;

    if(text_ptr > start_ptr) text_ptr++;
    lpDoc->selStart = (int)(text_ptr - start_ptr);

    priv_MapTextToScreen(lpDoc,lpDoc->selStart,&start_hilite);
    text_ptr = cur_ptr;
    while(!is_word_delimiter(*text_ptr) && text_ptr < end_ptr)
	text_ptr++;
    lpDoc->selEnd = (int)(text_ptr - start_ptr);  

    priv_MapTextToScreen(lpDoc,lpDoc->selEnd,&end_hilite);
    priv_DoHighlightRange(lpDoc,start_hilite,end_hilite);
    mem_UngraspHandle(hTxt);
    return;
}

/****************************************************************
		    ROUTINES TO HIGHLIGHT TEXT
*****************************************************************/

PRIVATE BOOL     hi_first_time;
PRIVATE Point    hi_far_snap;
PRIVATE Point    hi_near_snap;
PRIVATE int      hi_direction;

    /* values for hi_direction */
#define    TO_SOUTH 0
#define    TO_NORTH 1


/*---------------------------------------------------------*/
PRIVATE VOID  set_hi_far_pt(LPPOINT start_pt)
/*---------------------------------------------------------*/
{
    hi_far_snap   = *start_pt;
    hi_first_time = TRUE;
}

/*---------------------------------------------------------*/
PRIVATE VOID  priv_DoHighlightIncrementally(
    OPTIM LPDOC	lpDoc,
    LPPOINT	new_pt)
/*---------------------------------------------------------*/
{
    if(hi_first_time)
    {
	/* hi_far_snap was defined in start_highlight.
	 * Here we define hi_near_snap.
	 */
	hi_first_time = FALSE;
	hi_near_snap  = *new_pt;
	hi_direction  = (hi_far_snap.y <= hi_near_snap.y) 
				? TO_SOUTH : TO_NORTH;
	if(hi_direction==TO_SOUTH)
	{
	    priv_DoHighlightRange(lpDoc,hi_far_snap,hi_near_snap);
	}
	else /* TO_NORTH */
	{
	    priv_DoHighlightRange(lpDoc,
		hi_near_snap,hi_far_snap);	
	}
	return;
    }

    /* else: there is a range already highlighted */
    if(hi_direction==TO_SOUTH)
    {
	/* check for change in direction */
	if(new_pt->y < hi_far_snap.y
	    || (new_pt->y==hi_far_snap.y && new_pt->x < hi_far_snap.x)
	    )
	{

	    /* if so, unhighlight original and change direction */
	    priv_UnhighlightRange(lpDoc,hi_far_snap,hi_near_snap);
	    hi_direction = TO_NORTH;
	    hi_near_snap  = *new_pt;	/* structure copy */
	    priv_DoHighlightRange(lpDoc,hi_near_snap,hi_far_snap);
	}

	/* else check for un-highlighting a previously high region */
	else if(new_pt->y < hi_near_snap.y
	    || (new_pt->y==hi_near_snap.y && new_pt->x < hi_near_snap.x))
	{
	    priv_UnhighlightRange(lpDoc,/*pass*/new_pt,hi_near_snap);
	    hi_near_snap  = *new_pt;	/* structure copy */
	}

	/* else must just be extending the highlight in the same direction*/
	else 
	{
	    priv_DoHighlightRange(lpDoc,hi_near_snap,/*pass*/new_pt);
	    hi_near_snap  = *new_pt;	/* structure copy */
	}
    }
    else /* hi_direction==TO_NORTH*/
    {
	/* check for change in direction */
	if(new_pt->y > hi_far_snap.y
	   || (new_pt->y==hi_far_snap.y && new_pt->x > hi_far_snap.x)
	   )
	{
	    /* if so, unhighlight original and change direction */
	    priv_UnhighlightRange(lpDoc,hi_near_snap,hi_far_snap);
	    hi_direction = TO_SOUTH;
	    hi_near_snap  = *new_pt;	/* structure copy */
	    priv_DoHighlightRange(lpDoc,hi_far_snap,hi_near_snap);
	}

	/* else check for un-highlighting a previously high region */
	else if(new_pt->y > hi_near_snap.y
	    || (new_pt->y==hi_near_snap.y && new_pt->x > hi_near_snap.x)
	    )
	{
	    priv_UnhighlightRange(lpDoc,hi_near_snap,/*pass*/new_pt);
	    hi_near_snap  = *new_pt;	/* structure copy */
	}

	/* else must just be extending the highlight in the same direction */
	else
	{
	    priv_DoHighlightRange(lpDoc,/*pass*/new_pt,hi_near_snap);
	    hi_near_snap  = *new_pt;	/* structure copy */
	}
    }
}

/*---------------------------------------------------------*/
PRIVATE VOID	priv_DoHighlightRange(
    OPTIM LPDOC	lpDoc,
    LPPOINT	north_pt,
    LPPOINT	south_pt)
/*---------------------------------------------------------*/
{
    /* check for case of highlighting a single-line */
    if(north_pt->y == south_pt->y)
    {
	if(north_pt->x==south_pt->x)
	{
	    /*gp_InvertRect(north_pt->x,  north_pt->y,
		    south_pt->x + 1,
		    south_pt->y + lpDoc->lineHeight - 1);<<<don't highlight*/
	}
	else
	{
	    gp_InvertRect(north_pt->x,  north_pt->y,
		    south_pt->x , south_pt->y + lpDoc->lineHeight - 1);
	}
    }
    else    /* else:  must be multi-line highlight */
    {
	/* 1. first, do the northmost rect */
	gp_InvertRect(north_pt->x,    north_pt->y,
		    lpDoc->viewRect.right , north_pt->y + lpDoc->lineHeight - 1);

	/* 2. then, do middle rect (first check if there is a middle rect)*/
	if(north_pt->y + lpDoc->lineHeight < south_pt->y)
	{
	    int x1,y1,x2,y2;
	    for (  x1=lpDoc->viewRect.left, y1 = north_pt->y + lpDoc->lineHeight,
		   x2=lpDoc->viewRect.right, y2 = y1 + lpDoc->lineHeight - 1
		 ; y2 < south_pt->y
		 ; y1 += lpDoc->lineHeight, y2 = y1 + lpDoc->lineHeight - 1)
		 {
		     gp_InvertRect(x1,y1,x2,y2);
		 }

	    /*gp_InvertRect(lpDoc->viewRect.left, north_pt->y + lpDoc->lineHeight,
			lpDoc->viewRect.right,   south_pt->y - 1);*/
	}
		
	/* 3. lastly, do the southmost rect */	
	gp_InvertRect(lpDoc->viewRect.left, south_pt->y,
		     south_pt->x, south_pt->y + lpDoc->lineHeight - 1);
	    
    }
}


/****************************************************************/
PRIVATE int 	normalize_rects(LPDOC);
PRIVATE int  normalize_rects (OPTIM LPDOC lpDoc)
{
    int view_width = lpDoc->viewRect.right  - lpDoc->viewRect.left;
    int view_depth = lpDoc->viewRect.bottom - lpDoc->viewRect.top;
    int dest_depth = lpDoc->galleyRect.bottom - lpDoc->galleyRect.top;
    int dest_width;
	
    priv_ShowStructures(lpDoc);
    
    lpDoc->galleyRect.right = lpDoc->galleyRect.left 
				+ MAX(view_width, lpDoc->lineHeight * 2);
    dest_width = lpDoc->galleyRect.right  - lpDoc->galleyRect.left;	
    lpDoc->viewRect.right = lpDoc->viewRect.left + MAX(dest_width,view_width);
    if(dest_depth <= 0)
	lpDoc->galleyRect.bottom = lpDoc->galleyRect.top 
			+ MAX(view_depth, lpDoc->lineHeight);	    
    dest_depth = lpDoc->galleyRect.bottom - lpDoc->galleyRect.top;
    if(view_depth <= 0)
	    lpDoc->viewRect.bottom = lpDoc->viewRect.top + dest_depth;
    return dest_width;
}

/****************************************************************/
PRIVATE VOID priv_ReformatEntireDocument(LPDOC lpDoc)
{
    /* this reformats the entire text block */
    HMEMBLK	hTxt;
    int 	spec_measure;
    LPTEXTCH 	lpChar;
    LPTEXTCH 	end_position;
    int		num_phys_lines;
    int		cur_line;

    spec_measure = (lpDoc->galleyRect.right - lpDoc->galleyRect.left);

    if( spec_measure <= 0 
	|| lpDoc->galleyRect.top >= lpDoc->galleyRect.bottom) 
    {
	Warning("edRfmtEntireDocument: empty galleyRect!");
	spec_measure = normalize_rects(lpDoc);
    }


    priv_ResetXCoordCache();
    priv_RenewCharWidthCache(
	lpDoc->attrTypefont,
	lpDoc->attrTypeface,
	lpDoc->attrPointsize);

    hTxt = lpDoc->hTxt;
    lpTxt = mem_GraspHandle( hTxt);

    /*
     * The formatting of text consists of nested do-while loops.
     * The outer do-while calculates line breaks for a series of
     * lines in the text block, expanding the DOCSTRUC as necessary
     * to hold the lineStart[] data. The inner do-while calculates
     * the line break point for a single line, at the word space.
     */

    num_phys_lines = lpDoc->numPhysLines;
    cur_line = 0;
    lpDoc->lineMap[0] = 0;

    lpChar        = lpTxt;
    end_position = lpTxt + lpDoc->numTextChars;

    do 
    {
	 lpChar += priv_CalcLinebreak(lpChar, end_position, spec_measure);
	 
	 if(++cur_line >= num_phys_lines)
	 {
	     mem_UngraspHandle( (lpDoc->hLineMap));
	     lpDoc->hLineMap = mem_ChangeHandleSize( lpDoc->hLineMap,
		 (MEMSIZE) 
		 mem_GetHandleSize(lpDoc->hLineMap)
				 + (LINES_QUANTUM * sizeof(TEXTPOS)));
	     lpLineMap =  (LPLINEMAP) mem_GraspHandle(hDoc->hLineMap);
	     num_phys_lines += LINES_QUANTUM;
	 }
	 lpDoc->lineMap[cur_line] = (int)(lpChar - lpTxt);
    }  while  (lpChar < end_position);

    lpDoc->numPhysLines = num_phys_lines;
    lpDoc->numLines     = cur_line;
    lpDoc->lineMap[lpDoc->numLines] = lpDoc->numTextChars;

    /* galleyRect is "bottom-less", i.e., bottom grows to fit text */
    lpDoc->galleyRect.bottom = lpDoc->galleyRect.top + 
		    (lpDoc->numLines * lpDoc->lineHeight);

    mem_UngraspHandle( lpDoc->hLineMap);
    mem_UngraspHandle( hTxt);
}


/****************************************************************
 ED_FONT.C
****************************************************************/

/****************************************************************
	    FORMAT RECORD STRUCTURE
*****************************************************************/	    

typedef int POSITION;

typedef struct
{
    POSITION	position;
    BYTE	font;
    BYTE	face;
} FORMAT,
  FORMAT FAR * LPFORMAT;

#define    INVALID_POSITION	((POSITION) -1)
#define    INVALID_PARAM	((BYTE)     -1)


/*****************************************************************/

/****************************************************************/
PRIVATE LPFORMAT  ft_GetFirstFState(
    LPFORMAT  lpListStart, 
    int       text_pos,  
    LPBOOL  found)
{
#ifdef MULTI_FONT
    /* NOTE: this routine assumes that hFormat has been
     * locked by the caller. 
     * This returns the index of the first format record before
     * the specified text position. If none is found, the boolean
     * variable is set to FALSE.
     */
    OPTIM LPFORMAT lpFmt;
    
   for(   lpFmt = lpListStart
	; lpFmt->position != INVALID_POSITION && lpFmt->position <= text_pos
	; lpFmt++)
		/* continue */;
    if(lpFmt > lpListStart) 
    {
	*found = TRUE;
	return (LPFORMAT )(lpFmt-1);
    }
    else
    {
	*found = FALSE;
	return (LPFORMAT )lpListStart;
    }
    /* NOT REACHED*/
#else
    err_FatalError("multifont not implemented!");
#endif
}


/****************************************************************/
PRIVATE LPFORMAT   ft_GetNextFState(
    LPFORMAT pResumePoint, 
    int limit_pos, 
    LPBOOL found)
{
#ifdef MULTI_FONT
    /*
     * This routine returns a pointer to the next format record
     * just after the specified text position of the record passed in.
     * If none is found, the BOOL variable is set to FALSE.
     * Note: the resume_point is non-inclusive and the limit_pos
     * IS an inclusive position.
     */
    OPTIM LPFORMAT lpFmt = pResumePoint + 1;

    if(lpFmt->position != INVALID_POSITION && lpFmt->position <= limit_pos)
    {
	*found = TRUE;
	return lpFmt;
    }
    else
    {
	*found = FALSE;
	return pResumePoint;
    }
#else
    err_FatalError("multifont not implemented!");
#endif
}

/****************************************************************
    Each of the following functions manipulate the format record
    list. They generally deal with a selection-range (i.e., a
    start and end position in the text stream).
    
    There are various cases which they encounter, as follows.
    
    Case 0. Null range -- no work to do.
    Case 1. No format list present (handle is null) -- no work to do.
    Case 2. Minimal list (termination record only) -- no work to do.
    
    Case 3. Format records after range only (not before, not within).
	    Subtract delta only.
	    
    Case 4. Format records within range only (not before or after).
	    Remove range only.
	    
    Case 5. Format records within range and after (not before).
	    Remove range and subtract delta.

    Case 6. Format records before range only (not within and not after)
	    No work to do.
	    
    Case 7. Format records before range, and within range (not after)
	    Remove range only.
	    
    Case 8. Format records before range, and after, (not within)
	    Subtract delta only.
	    
    Case 9. Format records before range, within range, and after range.
	    Remove range and subtract delta.

    These cases are referred to by the above numbering in the code below.

****************************************************************/    


/****************************************************************/
PRIVATE VOID ft_DeleteFromFormat(
    LPDOC lpDoc,
    int range_start,
    int range_end)
{
#ifdef MULTI_FONT
    Size cur_size;
    int txt_bytes_deleted;
    OPTIM LPFORMAT lpFmt;	/* current pointer */
    OPTIM LPFORMAT lpListStart;		/* points to head of list */
    LPFORMAT pRangeStart;	/* points to just after range_start */
    LPFORMAT pRangeEnd;		/* points to just before range_end */
    
    /* Case 0. if range_start and range_end positions are identical, 
     * then there is nothing to delete 
     */
    txt_bytes_deleted = range_end - range_start;
    if(txt_bytes_deleted <= 0) return;

    /* Case 1. if no format handle, then there is nothing to delete */
    if(lpDoc->hFormat==NULL) 
    {
	err_FatalError("fmt:deletefromfmt"); /* was checked by caller*/
	return;
    }

    cur_size = mem_GetHandleSize( (lpDoc->hFormat)); 
    lpListStart = (LPFORMAT) mem_GraspHandle(lpDoc->hFormat); 

    /* Case 2. No format records in list.*/
    if(lpListStart->position==INVALID_POSITION) 
    {
	mem_UngraspHandle( (lpDoc->hFormat)); 
	return;
    }
    
    /* Case 3. Format records after range_end only */
    if(lpListStart->position > range_end)
    {
	for(  lpFmt = lpListStart
	    ; lpFmt->position != INVALID_POSITION
	    ; lpFmt++)
		lpFmt->position = MAX(0,(lpFmt->position - txt_bytes_deleted));
	mem_UngraspHandle( (lpDoc->hFormat)); 
	return;
    }

    /* either Case 4 or Case 5. (no records before range start). */
    if(lpListStart->position > range_start)
    {
	for(  lpFmt = lpListStart
	    ; lpFmt->position <= range_end && lpFmt->position != INVALID_POSITION
	    ; lpFmt++)
		/*continue*/;

	/* must be at least one record */
	if(lpFmt <= lpListStart) err_FatalError("deletefromformat #2");

	/* Case 4. Format records within range only (not before not after)*/
	if(lpFmt->position==INVALID_POSITION)
	{
	    /* structure copy of terminator block to head of list */
	    *lpListStart = *lpFmt;
	    mem_UngraspHandle( (lpDoc->hFormat)); 

	    /* remove range by truncating list to single terminator record */
	    lpDoc->hFormat = mem_ChangeHandleSize( lpDoc->hFormat,
		(MEMSIZE)  sizeof(FORMAT));
	    return;
	}

	pRangeEnd = lpFmt-1; /* save location just before range end */
	
	/* Case 5. Format records within range and after (not before)*/    

	/* record just before range end now becomes just AFTER range end */
	pRangeEnd->position = MIN(range_end+1,lpDoc->numTextChars);

	/* subtract delta from any records after and including range end */
	for(  lpFmt = pRangeEnd
	    ; lpFmt->position != INVALID_POSITION
	    ; lpFmt++)
		lpFmt->position = MAX(0,(lpFmt->position - txt_bytes_deleted));

	if(pRangeEnd > lpListStart)
	{
		/* delete range of records */
		misc_ShiftNBytes(
		    (MEMADDR)pRangeEnd,
		    (MEMADDR)lpListStart,
		    cur_size - ((int)pRangeEnd - (int)lpListStart));
	}
	mem_UngraspHandle( (lpDoc->hFormat)); 
	lpDoc->hFormat = mem_ChangeHandleSize( lpDoc->hFormat,
		    (MEMSIZE)  ((int)(pRangeEnd+1) - (int)lpListStart));
	return;
    }

    /* Cases 6,7, 8 or 9. */
    /* find the first format record BEFORE range_start position */
    for (   lpFmt = lpListStart
	  ; lpFmt->position < range_start 
	     && lpFmt->position != INVALID_POSITION
	  ; lpFmt++)
		    /*continue*/;

    /* Case 6. Format records before range only (not within and not after) */
    if(lpFmt->position==INVALID_POSITION)
    {
	mem_UngraspHandle( (lpDoc->hFormat)); 
	return;
    }
    
    /* must be at least one record before range */
    if(lpFmt <= lpListStart || lpFmt->position < range_start) 
		err_FatalError("deletefromformat");

    pRangeStart = lpFmt-1;

    for (
	  ; lpFmt->position <= range_end && lpFmt->position != INVALID_POSITION
	  ; lpFmt++)
		    /*continue*/;

    /* Case 7. Format records before and within only (not after) */
    if(lpFmt->position==INVALID_POSITION)	
    {
	if(lpFmt <= pRangeStart) err_FatalError("deletefromfmt #3");
	*(pRangeStart+1) = *lpFmt; /* structure copy of terminator block */
	mem_UngraspHandle( (lpDoc->hFormat)); 
	lpDoc->hFormat = mem_ChangeHandleSize( lpDoc->hFormat,
		    (MEMSIZE)  ((int)(pRangeStart+2) - (int)lpListStart));
	return;
    }

    if(lpFmt->position <= range_end) err_FatalError("deletefromfmt #4");
	
    /* Case 8. Format records before and after only (not within) */
    if(lpFmt==pRangeStart+1)    
    {
	for(
	    ; lpFmt->position != INVALID_POSITION
	    ; lpFmt++)
		lpFmt->position = MAX(0,(lpFmt->position - txt_bytes_deleted));

	mem_UngraspHandle( (lpDoc->hFormat)); 
	return;
    }


    /* Case 9. Format records before, within and after */
    pRangeEnd = lpFmt-1;

    if(pRangeEnd <= pRangeStart) err_FatalError("deletefromformat #5");
    
    pRangeEnd->position = MIN(range_end+1,lpDoc->numTextChars);

    for(  lpFmt = pRangeEnd
	; lpFmt->position != INVALID_POSITION
	; lpFmt++)
	    lpFmt->position = MAX(0,(lpFmt->position - txt_bytes_deleted));

     /* delete range of records */
     if(pRangeEnd > pRangeStart+1)
     {
	 misc_ShiftNBytes(
	     (MEMADDR)pRangeEnd,
	     (MEMADDR)(pRangeStart+1),
	     cur_size - ((int)pRangeEnd - (int)(lpListStart)));
     }
      mem_UngraspHandle( (lpDoc->hFormat)); 
      mem_ChangeHandleSize( (lpDoc->hFormat),
	    (MEMSIZE)  (cur_size - 
		    ((int)pRangeEnd - (int)(pRangeStart+1))));
      return;
#else
    err_FatalError("multifont not implemented!");
#endif
}

/****************************************************************/
PRIVATE VOID ft_AddToFormat(
    OPTIM LPDOC lpDoc,
    int where,
    int amount_added)
{
#ifdef MULTI_FONT
    LPFORMAT lpFmt;
    
    if( lpDoc->hFormat==NULL)
    {
	err_FatalError("fmt:addtofmt");
	return;
    }

    mem_GraspHandle( (lpDoc->hFormat)); 

    /* find the first format record after cursor position*/
    
    for(  lpFmt = (LPFORMAT ) *(lpDoc->hFormat)
	; lpFmt->position <= where && lpFmt->position != INVALID_POSITION
	; lpFmt++)
		/*continue */;

    if(lpFmt->position==INVALID_POSITION)
    {
        mem_UngraspHandle( lpDoc->hFormat); 
	return;
    }

    /* add one to subsequent format records */
    for(
	; lpFmt->position != INVALID_POSITION
	; lpFmt++)
	     lpFmt->position += amount_added;
    mem_UngraspHandle( lpDoc->hFormat); 
    return;
#else
    err_FatalError("multifont not implemented!");
#endif
}


/****************************************************************/
PRIVATE VOID  ft_NewFmtBlock(HDOC hDoc)
{
#ifdef MULTI_FONT
    HMEMBLK hFmt;
    LPFORMAT lpFmt;

    lpDoc->hFormat = hFmt = mem_AllocHandle((MEMSIZE)  sizeof(FORMAT));
    
    lpFmt = (LPFORMAT ) mem_GraspHandle(hFmt); 
    lpFmt->font     = INVALID_PARAM;
    lpFmt->face     = INVALID_PARAM;
    lpFmt->position = INVALID_POSITION;
    mem_UngraspHandle(hFmt);
#else
    err_FatalError("multifont not implemented!");
#endif
}
     

/****************************************************************/
PRIVATE VOID  ft_InsertFmtRecord(
    HMEMBLK hFmt,
    OPTIM LPDOC lpDoc,
    int font,
    int face,
    int cur_pos)
{
#ifdef MULTI_FONT
    LPFORMAT lpFmt,lpListStart,pJustAfter;
    Size cur_size;

    cur_size = mem_GetHandleSize(hFmt);
    lpDoc->hFormat = hFmt = 
	mem_ChangeHandleSize( hFmt,(MEMSIZE) (cur_size + sizeof(FORMAT)));
    
    lpListStart = (LPFORMAT ) mem_GraspHandle( hFmt);

    /* find the first format record after cursor position*/
    for(  lpFmt = lpListStart
	; lpFmt->position != INVALID_POSITION && lpFmt->position <= cur_pos
	; lpFmt++)
		/*continue*/;

    /* case: no format records after cursor position */
	    
    if(lpFmt->position==INVALID_POSITION)
    {
	LPFORMAT pAfter;
	
	/* no format record after cursor position */
	pAfter = lpFmt+1; /* we have increased the HMEMBLK size prior to this*/
	*pAfter = *lpFmt; /* structure copy of terminator record */
    }
    else
    {
	/* case: there are format records after cursor position */
	misc_ShiftNBytes(
	    (MEMADDR) lpFmt,
	    (MEMADDR) (lpFmt+1),
	    cur_size - ((int)lpFmt - (int)lpListStart));
    }
    /* if there was a prior record, copy it first, then modify */
    if(lpFmt > lpListStart)
    {
	    *lpFmt = *(lpFmt-1);
	    lpFmt->font = (font == INVALID_PARAM) ? (lpFmt->font) : (BYTE)font;
	    lpFmt->face = (face == INVALID_PARAM) ? (lpFmt->face) : (BYTE)face;
	    lpFmt->position = cur_pos;
     }
     else /* no prior record, so get information from HDOC record */
     {
	    lpFmt->font = (font == INVALID_PARAM) 
		? (lpDoc->attrTypefont) : (BYTE)font;
	    lpFmt->face = (face == INVALID_PARAM) 
		? (lpDoc->attrTypeface) : (BYTE)face;
	    lpFmt->position = cur_pos;
     }
    mem_UngraspHandle( lpDoc->hFormat); 
    return;
#else
    err_FatalError("multifont not implemented!emented");
#endif
}

/****************************************************************/
PRIVATE VOID  ft_ChangeParam(
    HDOC hDoc,
    int font,
    int face)
{
#ifdef MULTI_FONT
    LPDOC   lpDoc;
    HMEMBLK hFmt;

    if(lpDoc->hFormat==NULL) ft_NewFmtBlock(hDoc);

    lpDoc  = (LPDOC) mem_GraspHandle( hDoc);
    ft_PrintFmtList(lpDoc);

    hFmt = lpDoc->hFormat;

    if(lpDoc->selStart==lpDoc->selEnd)
    {
	ft_InsertFmtRecord(hFmt,lpDoc,font,face,lpDoc->selEnd);
    }
    else
    {
	FORMAT saved_format;
	OPTIM LPFORMAT lpListStart;
	OPTIM LPFORMAT lpFmt; 
	Size cur_size;
	
	cur_size = mem_GetHandleSize(hFmt);

	/* add space for two additional format records */
	lpDoc->hFormat = hFmt = mem_ChangeHandleSize(hFmt,
	    (MEMSIZE)  cur_size + (2*sizeof(FORMAT)));

	lpListStart = (LPFORMAT) mem_GraspHandle(hFmt);
	
	for( lpFmt = lpListStart
	    ; lpFmt->position < lpDoc->selStart 
	      && lpFmt->position != INVALID_POSITION
	    ; lpFmt++)
		/*continue*/;
	    
	/* get typographic state prior to selStart */
	if(lpFmt > lpListStart)
	    saved_format = *(lpFmt-1);
	else
	{
	    saved_format.face = lpDoc->attrTypeface;
	    saved_format.font = lpDoc->attrTypefont;
	}

	/* case: no format records after selStart */    
	if(lpFmt->position==INVALID_POSITION)    
	{
	    LPFORMAT pAfter = lpFmt + 2;
	    *pAfter = *lpFmt; /* structure copy of termination record */

	    lpFmt->face = (face==INVALID_PARAM? saved_format.face : face);
	    lpFmt->font = (font==INVALID_PARAM? saved_format.font : font);
	    lpFmt->position = lpDoc->selStart;
	    lpFmt++;
	    lpFmt->face = saved_format.face;
	    lpFmt->font = saved_format.font;
	    lpFmt->position = lpDoc->selEnd;

	    mem_UngraspHandle(hFmt);
	    ft_PrintFmtList(lpDoc);
	    return;
	}

	/* else must be format records after selStart */
	misc_ShiftNBytes(
	    (MEMADDR) lpFmt,
	    (MEMADDR) (lpFmt+1),
	    cur_size - ((int)lpFmt - (int)lpListStart));

	lpFmt->face = (face==INVALID_PARAM? saved_format.face : face);
	lpFmt->font = (font==INVALID_PARAM? saved_format.font : font);
	lpFmt->position = lpDoc->selStart;

	for(
	    ; lpFmt->position <= lpDoc->selEnd
	      && lpFmt->position != INVALID_POSITION
	    ; lpFmt++)
		/*continue*/;

	/* case: no format records after selEnd */    
	if(lpFmt->position==INVALID_POSITION)    
	{
	    LPFORMAT pAfter = lpFmt + 1;
	    *pAfter = *lpFmt; /* structure copy of termination record */
	}
	else
	{
	    /* else there are format records after selEnd */
	    misc_ShiftNBytes(
		(MEMADDR) lpFmt,
		(MEMADDR) (lpFmt+1),
		cur_size + sizeof(FORMAT) - ((int)lpFmt - (int)lpListStart));
	 }
	 lpFmt->face    = saved_format.face;
	 lpFmt->font    = saved_format.font;
	 lpFmt->position= lpDoc->selEnd;
	    
	 mem_UngraspHandle(hFmt);
     }
    ft_PrintFmtList(lpDoc);
    return;
#else
    err_FatalError("multifont not implemented!");
#endif
 }

/****************************************************************/
PRIVATE VOID ft_DrawFmtLine(
    LPDOC	lpDoc,
    HMEMBLK	hTxt,
    int		cur_line,
    int		y)
{
#ifdef MULTI_FONT
    /* this routine is only called if format list is NOT NULL */
	
    OPTIM LPFORMAT lpFmt;
    LPFORMAT pFormatList;
    LPXCOORD lpCoord;
    LPTEXTCH lpTxt;
    int start_pos,end_pos,cur_pos,cur_end;
    int x,i;
    BOOL found;
    HMEMBLK hLine;
    
    lpTxt = (LPTEXTCH) mem_GraspHandle(lpDoc->hTxt);
    start_pos  = lpDoc->lineMap[cur_line];
    end_pos    = lpDoc->lineMap[cur_line+1];

    /* the following disables NEWLINE char at end of line */
    if( lpTxt[end_pos-1]==NEWLINE_CH)
    {
	end_pos--;
    }

    /* priv_GetLineHandleXCoord also goes through a ft_GetFirstFState()
     * and ft_GetNextFState() sequence, and makes gp_ChangeTypeface, and
     * gp_ChangeTypefont calls to GraphPrim . Therefore it should be called
     * before our own sequence, so that the GraphPrim Text state
     * is what we think and is not changed underneath by subordinate
     * routines.
     * Also it locks hFormat, so it should be called before call to
     * mem_GraspHandle(lpDoc->hFormat).
     */

    hLine = priv_GetLineHandleXCoord(lpDoc, lpTxt,cur_line);

    lpCoord = (LPXCOORD ) mem_GraspHandle(hLine);

    lpCoord++;/* skip the first element of lpCoord array */

    pFormatList = (LPFORMAT ) mem_GraspHandle(lpDoc->hFormat);

    cur_pos = start_pos;
    cur_end = end_pos;
    found = FALSE;
    ft_PrintFmtList(lpDoc);
    lpFmt =  ft_GetFirstFState(pFormatList,cur_pos, &found);

    gp_SetTextMode(lpDoc->attrMode);    
    gp_ChangePointsize(lpDoc->attrPointsize);
    if(found)   
    {
	gp_ChangeTypeface(lpFmt->face);  
	gp_ChangeTypefont(lpFmt->font ); 
	lpFmt = ft_GetNextFState(lpFmt, end_pos, &found);	    
    }
    else	
    {
	gp_ChangeTypeface(lpDoc->attrTypeface); 
	gp_ChangeTypefont(lpDoc->attrTypefont); 
	lpFmt = ft_GetNextFState((pFormatList-1), end_pos, &found);
    } 

    while( cur_pos < end_pos)
    {
	cur_end = (found && lpFmt->position != INVALID_POSITION)	    
	          ? MIN(end_pos,lpFmt->position)
		  : end_pos;

	gp_TextOutAt( *(lpCoord + (cur_pos - start_pos)), /* x-coord */
		     y,					/* y-coord */
		     lpDoc,
		     lpDoc->hTxt,
		     cur_pos,cur_end-cur_pos);

	if(found)
	{
	    gp_ChangeTypeface(lpFmt->face); 
	    gp_ChangeTypefont(lpFmt->font);
	}
	lpFmt =  ft_GetNextFState(
		    (LPFORMAT )(found? lpFmt : pFormatList - 1), 
			   end_pos, &found);	    
	cur_pos = cur_end;
    }

    mem_UngraspHandle(hLine);
    mem_UngraspHandle(lpDoc->hFormat);
    return;
#else
    err_FatalError("multifont not implemented!");
#endif
}


/****************************************************************/
PRIVATE VOID ft_PrintFmtList(LPDOC lpDoc)
{
#ifdef MULTI_FONT
    int       i;
    HMEMBLK    hFmt;
    LPFORMAT  lpFmt,lpListStart;
    if(lpDoc->hFormat==NULL)
    {
	return;
    }
    hFmt = lpDoc->hFormat;
    lpListStart = (LPFORMAT )  mem_GraspHandle(hFmt);  
    for(  lpFmt = lpListStart,i=0
	; lpFmt->position != INVALID_POSITION && i < 20
	; lpFmt++,i++)
	{
	    /*show structures here*/;
	}
    mem_UngraspHandle(hFmt);
#else
    err_FatalError("multifont not implemented!");
#endif /*MULTI_FONT*/
}


/****************************************************************
	PRIVATELY VISIBLE TEXTEDIT EXTENSIONS
****************************************************************/

PUBLIC VOID ed_ChangeFont(
    HDOC hDoc,
    int f)
{
#ifdef MULTI_FONT
    LPDOC lpDoc;
    ft_ChangeParam(hDoc,f,INVALID_PARAM);
    lpDoc = (LPDOC) mem_GraspHandle( hDoc); 
    priv_UpdateView(&(lpDoc->viewRect),hDoc);
    mem_UngraspHandle( hDoc);
#else
    lpDoc->attrTypefont = f;
#endif
}

/****************************************************************/
PUBLIC VOID ed_ChangeFace(hDoc,f)
    HDOC hDoc;
    int f;
{
#ifdef MULTI_FONT
    LPDOC lpDoc;
    ft_ChangeParam(hDoc,INVALID_PARAM,f);
    lpDoc = (LPDOC) mem_GraspHandle( hDoc);
    priv_UpdateView(&(lpDoc->viewRect),hDoc);
    mem_UngraspHandle( hDoc);
#else
    lpDoc->attrTypeface = f;
#endif
}

/****************************************************************/
PUBLIC VOID ed_ChangeSize(hDoc,s)
    HDOC hDoc;
    int s;
{
#ifdef MULTI_FONT
    /*not yet implemented*/;
#else
    lpDoc->attrTypefont = s;
#endif
}

/****************************************************************
	 UTILITY FUNCTIONS
****************************************************************/

PRIVATE VOID  misc_ShiftNBytes(
    OPTIM MEMADDR from,
    OPTIM MEMADDR to,
    OPTIM int len)
{
    if (from>to)
		while (len--) *to++ = *from++;
    else	
    {
		from += len;
		to   += len;
		while (len--) *--to = *--from;
    }
}


/****************************************************************/

PRIVATE int  multiplied_amount	(int);

#ifdef OUTPUT_BITMAP_CACHE
PRIVATE VOID  priv_InitBitmapCache	(void);
PRIVATE VOID  priv_ResetBitmapCache	(void);
PRIVATE VOID  priv_FiniBitmapCache	(void);
PRIVATE VOID  priv_CacheOneBitmap	(int,int,int,int);
PRIVATE VOID  priv_DrawOneBitmap	(int,int,int,int);
#endif


/*PRIVATE  VOID  priv_BuildCachedLine	(LPDOC, LPTEXTCH,int,HMEMBLK);*/

/****************************************************************
	TEXT TO SCREEN AND CONVERSE
****************************************************************/

/*--------------------------------------------------------*/
PRIVATE VOID  priv_MapScreenToText(
    LPDOC	lpDoc,
    LPPOINT 	screen_pt,
    LPINT	text_pos,
    LPPOINT	snapped_pt)
/*--------------------------------------------------------*/
{
    /* this routine is called from ed_OnMouseDown() and ed_OnDoubleClick().
     * Given a screen coordinate, map that location to a position within
     * the text stream. Snap that text position to the limits of the text
     * stream (e.g., last char in the stream). Also return the corresponding
     * screen point (i.e., constrained to the character grid). This position
     * is also snapped to the limits of the viewRect
     */
    int		y;
    int		cur_line;

    /*------------ first, find the y-coordiante --------*/

    for (  cur_line = 0, y = lpDoc->galleyRect.top
	 ; cur_line < lpDoc->numLines && y <= screen_pt->y
         ; cur_line++, y += lpDoc->lineHeight    
	)
    {
	/* continue */;
    }

    /* then back up one if not at first position */
    if(cur_line > 0)
    {
	cur_line--;
	snapped_pt->y = y - lpDoc->lineHeight;
    }
    else
	snapped_pt->y = y;

    /*------------ then, find the x-coordiante --------*/
    BLOCK
    {
	OPTIM LPXCOORD lpCoord;
	OPTIM LPXCOORD pEnd;
	HMEMBLK	hLine;
	int	cur_pos,end_pos,start_pos,num_chars;

	start_pos = lpDoc->lineMap[cur_line];
	end_pos   = 
	    (cur_line== lpDoc->numLines) /* the line after last line */
			? lpDoc->numTextChars
			: lpDoc->lineMap[cur_line + 1];    
	num_chars = end_pos - start_pos;	    
	hLine     = priv_GetLineHandleXCoord(lpDoc, lpTxt,cur_line);
	lpCoord    = (LPXCOORD) mem_GraspHandle(hLine);
	lpCoord++; /* skip the zeroth element in array (contains linewidth)*/

	for ( pEnd = lpCoord + num_chars, cur_pos = start_pos
	    ; lpCoord <= pEnd && *lpCoord <= screen_pt->x
	    ; lpCoord++, cur_pos++
	    )
		/*continue*/;

	/* then, back up one if necessary */
	if( lpCoord <= pEnd)
	{
	    if(cur_pos > start_pos)
		{ --cur_pos; --lpCoord; }
					    /* is this MIN needed? */
	    *text_pos = MIN(cur_pos,lpDoc->numTextChars); 
	    snapped_pt->x = *lpCoord;
	    mem_UngraspHandle(hLine);
	}
	else /* passed the very last char on this line, go to next */
	{
	    *text_pos = end_pos;

	    if(end_pos < lpDoc->numTextChars) /* if not the last character */
	    {
		snapped_pt->y += lpDoc->lineHeight;
		cur_line++;
		mem_UngraspHandle(hLine); 

		/* unlock BEFORE priv_GetLineHandleXCoord */

		snapped_pt->x= 
	    *((LPXCOORD ) *(priv_GetLineHandleXCoord(lpDoc,lpTxt,cur_line))
				  + 1);
	    }
	    else  /* yes it is the last character, AND it is a newline */
		if( *(((LPTEXTCH )*(lpDoc->hTxt)) + end_pos) == NEWLINE_CH)
	    {
		snapped_pt->y += lpDoc->lineHeight;
		cur_line++;
				/* unlock BEFORE priv_GetLineHandleXCoord */
		mem_UngraspMem(hLine); 
		snapped_pt->x= 
		    *((LPXCOORD ) *(priv_GetLineHandleXCoord(lpDoc,lpTxt,cur_line))
				  + 1);
	    }
	    else /* it is the last character, and it is NOT a newline */
	    {
	    	snapped_pt->x = *(--lpCoord);
		mem_UngraspHandle(hLine);
	    }
	}
    }
    /* constrain the already snapped point to the viewRect boundary */
    if(snapped_pt->y < lpDoc->viewRect.top)
    {
	int dy = ((lpDoc->viewRect.top - lpDoc->galleyRect.top) 
						% lpDoc->lineHeight);


	snapped_pt->y = dy 
			? lpDoc->viewRect.top - dy + lpDoc->lineHeight
			: lpDoc->viewRect.top; 
	snapped_pt->x = MAX(lpDoc->viewRect.left,lpDoc->galleyRect.left);

    }
    else
    {
	int bottom = lpDoc->viewRect.bottom 
			- ((lpDoc->viewRect.bottom - lpDoc->galleyRect.top) 
						 % lpDoc->lineHeight)
		        - lpDoc->lineHeight;

	if( snapped_pt->y > bottom)
	{
	    snapped_pt->y = bottom;
	    snapped_pt->x = 
		MIN(lpDoc->viewRect.right,lpDoc->galleyRect.right);
	}
	else
	{
	    if(snapped_pt->x < lpDoc->viewRect.left) 
	    {
		snapped_pt->x = 
		    MAX(lpDoc->viewRect.left,lpDoc->galleyRect.left);
	    }
	    else if (snapped_pt->x > lpDoc->viewRect.right) 
	    {
		snapped_pt->x = 
		    MIN(lpDoc->viewRect.right,lpDoc->galleyRect.right);
	    }
	 }
    }
}

/*--------------------------------------------------------*/
PRIVATE BOOL  priv_MapTextToScreen(
    OPTIM LPDOC	lpDoc,
    int		text_pos,
    LPPOINT	screen_pt)
/*--------------------------------------------------------*/
{
    /* NOTE: this function assumes lineMap[] is valid */
    int		cur_line;
    int		snapped_view_bott;
    int		x,y;
    LPTEXTCH	lpTxt;
    
    lpTxt = (LPTEXTCH) mem_GraspHandle(lpDoc->hTxt);	

    /*------------first, find y-coordinate--------------*/
		
    /* check case of last character */
    if(text_pos==lpDoc->numTextChars)
    {
	cur_line = MAX(lpDoc->numLines-1,0);
	y = lpDoc->galleyRect.top + (cur_line * lpDoc->lineHeight);
    }
    else
    {
	/* else, find y-coordinate by stepping through galleyRect grid */
	for ( cur_line = 0, y = lpDoc->galleyRect.top
	    ; cur_line < lpDoc->numLines 
	      && lpDoc->lineMap[cur_line+1] <= text_pos
	    ; cur_line++, y += lpDoc->lineHeight    
	    )
	    { /* continue */; }
    }
	/* position cursor on the next line after a NL */
    if(    text_pos > 0
	&& lpTxt[text_pos-1] == NEWLINE_CH
	&& text_pos >= lpDoc->numTextChars) /*>>>should be LESS THAN! <<<*/
    {
	y += lpDoc->lineHeight; cur_line++;
    }
    snapped_view_bott =  lpDoc->viewRect.bottom 
	- ((lpDoc->viewRect.bottom - lpDoc->galleyRect.top) 
	    % lpDoc->lineHeight)
	- lpDoc->lineHeight;

    /* then, snap y to viewRect grid */
    if(y < lpDoc->viewRect.top)
    {
	int dy = ((lpDoc->viewRect.top - lpDoc->galleyRect.top) 
			% lpDoc->lineHeight);
	screen_pt->y = dy 
			? lpDoc->viewRect.top - dy + lpDoc->lineHeight
			: lpDoc->viewRect.top; 
	screen_pt->x = MAX(lpDoc->viewRect.left,lpDoc->galleyRect.left);

	return FALSE;
    }
    else if (y > snapped_view_bott)
    {
	screen_pt->y = snapped_view_bott;
	screen_pt->x = MIN(lpDoc->viewRect.right,lpDoc->galleyRect.right);    
	return FALSE;
    }

    /* else, we are within viewRect, so press on */
    screen_pt->y = y;

    /*----------- then, find x-coordinate ------------*/
    BLOCK
    {
	HMEMBLK	hLine;
	
	hLine     = priv_GetLineHandleXCoord(lpDoc,lpTxt, cur_line);

	/* index into array equals:  text_pos - lpDoc->lineMap[cur_line] */
	/* add one to index to skip over line-width (zeroth array element) */
	
	x = *(LPXCOORD )( ((LPXCOORD ) *hLine) 
				+ text_pos - lpDoc->lineMap[cur_line] + 1);
    }

    if( x < lpDoc->viewRect.left)
    {
	screen_pt->x = lpDoc->viewRect.left;
	return FALSE;
    }
    else if (x > lpDoc->viewRect.right)
    {
	screen_pt->x = lpDoc->viewRect.right; /*not quite correct */
	return FALSE;
    }
    /* else */
    screen_pt->x = x;
    return TRUE;
}


/*--------------------------------------------------------*/
PRIVATE int 
priv_LocateLine(
    OPTIM LPDOC lpDoc, 
    int pos
    )
/*--------------------------------------------------------*/

{
    /* NOTE: this assumes lineMap are valid
     * and that the HDOC record pointed to by hDoc has been locked
     */
    OPTIM int cur_line;
    
    for ( cur_line = 0
	; cur_line < lpDoc->numLines 
	; cur_line++
	)
    {
	if (lpDoc->lineMap[cur_line + 1] > pos) break;
    }
    return cur_line;
}


/****************************************************************
		CARET MANAGER
*****************************************************************/

PRIVATE int   prev_caret_height;


/*--------------------------------------------------------*/
PRIVATE VOID	  priv_InitCaret(void)
/*--------------------------------------------------------*/
{
    ;
}


/*--------------------------------------------------------*/
PRIVATE VOID 	  priv_DestroyCaret(void)
/*--------------------------------------------------------*/
{
    DestroyCaret();
    prev_caret_height = INVALID_COORD;
}


/*--------------------------------------------------------*/
PRIVATE VOID  priv_DrawCaret(
	    LPPOINT	caret_pt,
	    OPTIM LPDOC	lpDoc)
	    )
/*--------------------------------------------------------*/
{
    HWND hWnd = lpDoc->hWnd;

    if(prev_caret_height != lpDoc->lineHeight)
    {
	DestroyCaret();
	CreateCaret(hWnd,NULL,0,lpDoc->lineHeight);
	prev_caret_height = lpDoc->lineHeight;
    }  
    SetCaretPos(caret_pt->x,caret_pt->y);
    ShowCaret(hWnd);
}


/****************************************************************
  INTERFACE TO ENVIRONMENT (interface to graphics primitives)
*****************************************************************/

/*--------------------------------------------------------*/
PRIVATE VOID  gp_InvertRect(
    int x1, int y1,
    int x2, int y2)
/*--------------------------------------------------------*/
{
    RECT r;
    r.left = x1; r.top = y1;
    r.right = x2; r.bottom = y2;
    InvertRect(&r);
}

/*--------------------------------------------------------*/
PRIVATE VOID  gp_ClearRect(
    LPDOC   lpDoc,
    int x1, int y1,
    int x2, int d
    )	
/*--------------------------------------------------------*/
{
    /* Note: the arguments to this function describe a rectangle,
     * but in a slightly odd way, because it makes it easier
     * for the calling function.
     * Rather than using (x,y,width,depth) or using (x1,y1,x2,y2), 
     * this function uses a mix of the above calling protocols:
     * (x1,y1,x2,depth).
     */
    RECT r;
    HDC hDC;
    
    r.left  = x1; 	r.top    = y1;
    r.right = x2;	r.bottom = y1+d;

    hDC = win_GetDC(lpDoc->hWnd);
    win_SelectObject(hDC,GetStockObject(WHITE_BRUSH));        
    PatBlt(hDC, r.left, r.top, r.right - r.left,
		    r.bottom - r.top, PATCOPY);
    win_ReleaseDC(lpDoc->hWnd,hDC);	
}
    

/*---------------------------------------------------------*/
PRIVATE VOID gp_TextOutAt(x,y,lpDoc,hTxt,start_pos,count)
	    int		x,y; 
	    OPTIM LPDOC	lpDoc; 
	    HMEMBLK	hTxt; 
	    int		start_pos,count;
/*--------------------------------------------------------*/
{
#ifdef WINDOWS_SPECIFIC_CODE
#	define    BLACK 0L
#	define    WHITE 0x00FFFFFFL
    HWND hWnd = lpDoc->hWnd;
    HDC hDC;
    hDC = GetDC(hWnd);
    if(hDC)
    {
	SetTextColor(hDC, BLACK); 
	SetBkColor(hDC,WHITE);
	SetBkMode(hDC,TRANSPARENT);
	TextOut(hDC, x, y, 
	     &(lpTxt[start_pos]), count);
    }
    ReleaseDC(hWnd,hDC);
#endif
}


/*---------------------------------------------------------*/
PRIVATE BOOL  kb_IsThereKeystroke(OPTIM LPDOC lpDoc)
/*---------------------------------------------------------*/
{
    MSG msg;
    HWND hWnd = lpDoc->hWnd;

    if (PeekMessage((LPMSG)&msg, hWnd, WM_KEYFIRST,WM_KEYLAST,
		     FALSE))	/* don't remove from queue */
    {
	return (msg.message == WM_CHAR || msg.message==WM_KEYDOWN);
    }
    return FALSE;
}

/****************************************************************
	CLIPPING
****************************************************************/

/*---------------------------------------------------------*/
PRIVATE VOID	gp_SetClipRect(
    HDOC	hDoc,
    LPRECT	r)
/*---------------------------------------------------------*/
{
    long window_origin;
    HWND hWnd;
    HDC hDC;
    HRGN temp_clip_region;
    HANDLE h;

    hWnd = lpDoc->hWnd;
    hDC = win_GetDC(hWnd,hDC);

    temp_clip_region = CreateRectRgn(r->left,r->top,r->right,r->bottom);
    window_origin = GetWindowOrg(hDC);
    OffsetRgn(temp_clip_region, 
		- LOWORD(window_origin), 
		- HIWORD(window_origin));
    SaveDC(hDC);
    h = win_SelectObject(hDC,temp_clip_region);
    win_ReleaseDC(hWnd,hDC); /*<<<<actually, this should happen later!<<*/
}

/*---------------------------------------------------------*/
PRIVATE VOID	gp_RestoreClipRect(HDOC hDoc)
/*---------------------------------------------------------*/
{
    RestoreDC(theCurrentDC, -1);
}

/*---------------------------------------------------------*/
PRIVATE VOID	gp_ScrollWindow(lpDoc,dh,dv,rect)
    LPDOC	lpDoc;
    int		dh,dv;
    LPRECT	rect;
/*---------------------------------------------------------*/
{
    if(    rect->right   > rect->left
        && rect->bottom  > rect->top)
    {
	win_ScrollWindow(
		lpDoc->hWnd,
		dh,dv,	
		(LPRECT) ms_rect,
		(LPRECT) NULL);
    }
}


/****************************************************************
	    DRAW A LINE OF TEXT
******************************************************************/

/*---------------------------------------------------------*/
PRIVATE VOID  gp_DrawLineOfText (
    OPTIM LPDOC	    lpDoc,
    OPTIM LPTEXTCH  hTxt,
    int		cur_line,
    int		y,
    BOOL	use_bitmap)
/*---------------------------------------------------------*/
{
    HMEMBLK	hXcoord;
    LPXCOORD	lpXcoord;
    int start_pos, count, x;

#ifdef MULTI_FONT
    if(lpDoc->hFormat)
    {
	extern void ft_DrawFmtLine();
	ft_DrawFmtLine (lpDoc, hTxt, cur_line, y); 
	return;
    }
#endif

#ifdef RV /* compensate for difference in where char is drawn on scrn */
#else
    y += lpDoc->lineHeight - (lpDoc->lineHeight >> 2);
#endif
    start_pos  = lpDoc->lineMap[cur_line];
    count      = lpDoc->lineMap[cur_line+1] - start_pos;
    hXcoord    = priv_GetLineHandleXCoord(lpDoc,lpTxt, cur_line);

    /* the following line disables drawing of newline char */
    if(lpTxt[start_pos+count-1]==NEWLINE_CH) count--;

    lpXcoord = (LPXCOORD) mem_GraspHandle(hXcoord);
    x = lpXcoord[1];

#ifdef OUTPUT_BITMAP_CACHE
    if(!use_bitmap)
    {
	gp_TextOutAt(x,y,lpDoc,hTxt,start_pos,count);
    }
    else
    {
	int line_width = *((LPXCOORD ) *hXcoord);

	if(!is_bitmap_valid)
	{
	    gp_TextOutAt(x,y,lpDoc,hTxt,start_pos,count);
	    priv_CacheOneBitmap(x,y,line_width,lpDoc->lineHeight);
	    is_bitmap_valid = TRUE;
	}
	else
	{
	    priv_DrawOneBitmap(x,y,line_width,lpDoc->lineHeight);
	}
    }
#else
    gp_TextOutAt(x,y,lpDoc,hTxt,start_pos,count);
#endif
}

#ifdef EMACS_COMMANDS
/****************************************************************
		    EMACS_COMMANDS
*****************************************************************/
PRIVATE int cmd_multiplier;

/*---------------------------------------------------------*/
PRIVATE int  multiplied_amount(int m)
/*---------------------------------------------------------*/
{
    int seed = 4;
    while (m--)
	seed *= seed;
    return seed;
}

/*-----------------------------------------------------------*/
PRIVATE VOID priv_DoEmacsCommand(
    TEXTCH cmd,
    LPDOC lpDoc)
/*-----------------------------------------------------------*/
{
    OPTIM LPDOC lpDoc;
    int amount;

    if(cmd != CTL('L') && lpDoc->selEnd != lpDoc->selStart)
    {
	/* ignore emacs commands when highlighted, for now, except CTL-L */
	return;
    }
    priv_UnshowSelection(lpDoc);
    switch(cmd)
    {
	case CTL('A'):	lpDoc->selEnd = 0;		
			break;

	case CTL('B'):	amount = (cmd_multiplier > 0)
					? multiplied_amount(cmd_multiplier)
					: 1;
			lpDoc->selEnd = (lpDoc->selEnd - amount > 0)
					? lpDoc->selEnd - amount
					: 0; 
			break;
			
	case CTL('D'):	priv_OnCharMsg( CTL('F'),     hDoc);
			priv_OnCharMsg( BACKSPACE_CH, hDoc);
			return;
	    
	case CTL('E'):	lpDoc->selEnd = lpDoc->numTextChars;	
			break;	

	case CTL('F'):	amount = (cmd_multiplier > 0)
					? multiplied_amount(cmd_multiplier)
					: 1;

			lpDoc->selEnd = 
			    (lpDoc->selEnd + amount < lpDoc->numTextChars)
					 ? lpDoc->selEnd + amount
					 : lpDoc->numTextChars; 
			break;		 
			
	case CTL('K'):	lpDoc->numTextChars = lpDoc->selEnd;
			priv_ReformatEntireDocument(lpDoc);
			priv_UpdateView(&(lpDoc->viewRect),hDoc);
			return;

	case CTL('L'):	
			priv_ReformatEntireDocument(lpDoc);
			priv_UpdateView(&(lpDoc->viewRect),lpDoc);
			return;

	case CTL('N'):	{
			int cur_line, next_line, line_after;
			int start_pos, end_pos;
	
			if(lpDoc->selEnd == lpDoc->numTextChars) 
			    break;
			
			amount = (cmd_multiplier > 0)
					? multiplied_amount(cmd_multiplier)
					: 1;

			cur_line  = priv_LocateLine(lpDoc,lpDoc->selEnd);
			next_line = cur_line + amount;
			if( next_line >= lpDoc->numLines)
			{
			    lpDoc->selEnd = lpDoc->numTextChars;
			    break;
			}
			/* else */
			line_after = next_line + 1;
			if( line_after >= lpDoc->numLines)
			    end_pos = lpDoc->numTextChars;
			else
			    end_pos = lpDoc->lineMap[line_after] - 1;
			start_pos = lpDoc->lineMap[next_line];

			lpDoc->selEnd =  MIN(start_pos 
			    + (lpDoc->selEnd - lpDoc->lineMap[cur_line]), 
			    end_pos);
			}
			break;	    

	case CTL('P'):	
			{
			int cur_line, prev_line;
			int start_pos, end_pos;
	
			if(lpDoc->selEnd == 0)  break;

			amount = (cmd_multiplier > 0)
					? multiplied_amount(cmd_multiplier)
					: 1;

			cur_line   = priv_LocateLine(lpDoc,lpDoc->selEnd);
			prev_line  = cur_line - amount;
			if(prev_line < 0)
			{
			    lpDoc->selEnd = 0;
			    break;
			}
			/*else*/
			start_pos = lpDoc->lineMap[prev_line];
			end_pos   = lpDoc->lineMap[cur_line ]-1;
			
			lpDoc->selEnd = MIN(start_pos + 
			    (lpDoc->selEnd - lpDoc->lineMap[cur_line]),
			    end_pos);
			}
			break;	    

	case CTL('U'):	cmd_multiplier++;
			return;
	
	case CTL('V'):  priv_ShowStructures(lpDoc);
			priv_ShowSelection(lpDoc);
			return;	
    }
    cmd_multiplier = 0;
    lpDoc->selStart = lpDoc->selEnd;
    priv_ShowSelection(lpDoc);
    return;
}
#endif /*EMACS_COMMANDS*/



#ifdef  CHAR_WIDTH_CACHE
/****************************************************************
CHAR WIDTH CACHE --- cached on a font/face/size combination basis
****************************************************************/	


#ifdef STANDALONE_TEST /*12/14*/
PRIVATE BOOL do_wind_font;
PRIVATE BOOL DoingShowPage;
#else
IMPORT BOOL do_wind_font;
IMPORT BOOL DoingShowPage;
#endif


#define    SIZEOF_WIDTH_CACHE 4
#define    MAX_CHAR_NUM       255

typedef struct
{
    int 	font;
    int		face;
    int		size;
    BOOL 	in_use;
    int		widths[MAX_CHAR_NUM+1];
} WREC;

#define    INVALID_FONT	(-1)
#define    INVALID_FACE    (-1)


PRIVATE WREC  *cached_widths;
PRIVATE int   *cur_width_table;

#define    char_width(i)		(cur_width_table[(i)])

/****************************************************************/
PRIVATE VOID priv_InitCharWidthCache()
{
    OPTIM WREC *p;
    
    cached_widths = (WREC *)mem_NewPtr((MEMSIZE) (sizeof(WREC)*SIZEOF_WIDTH_CACHE));
    if(cached_widths==NULL) err_FatalError("fmt: initcharwidthcache");

    for(  p = cached_widths 
	; p < &cached_widths[SIZEOF_WIDTH_CACHE]
	; p++)
    {
	p->face		= INVALID_FACE;
	p->font		= INVALID_FONT;
	p->in_use	= FALSE;
    }
    cur_width_table = NULL;
}


/****************************************************************/
PRIVATE VOID priv_RenewCharWidthCache(font,face,size)
    int font,face,size;
{
    OPTIM WREC *p;
    BOOL found;
    
    
    for(  p = cached_widths 
	; p < &cached_widths[SIZEOF_WIDTH_CACHE]
	; p++)
    {
	if(    p->in_use
	    && p->font  == font
	    && p->face  == face
	    && p->size  == size)
	    {
		extern DISPLAY theDisplay;
		if( theDisplay->attrTypefont != font )	
		    gp_ChangeTypefont(font);
		if( theDisplay->attrTypeface != face )	
		    gp_ChangeTypeface(face);
		if( theDisplay->attrPointsize != size )	
		    gp_ChangePointsize(size);
	    }
	    cur_width_table = &(p->widths[0]);
	    return;
	}
    }
    /* else */
    for(  p = cached_widths, found = FALSE
	; p < &cached_widths[SIZEOF_WIDTH_CACHE]
	; p++)
    {
	if(!p->in_use)
	{
		p->in_use = TRUE; found = TRUE; break;
	}
    }

    if(!found)	/* a free slot was not found, re-use the first slot */
    {
	p = cached_widths;
    }

	do_wind_font = FALSE;
    p->font = font;    gp_ChangeTypefont(font);    
    p->face = face;    gp_ChangeTypeface(face);    
	do_wind_font = TRUE;
    p->size = size;    gp_ChangePointsize(size);    
    cur_width_table = &(p->widths[0]);

    BLOCK
    {
	OPTIM int i   = MAX_CHAR_NUM+1;
	OPTIM LPINT cp = &(p->widths[MAX_CHAR_NUM]);
	while(i--) *cp-- = gp_CharWidth((TEXTCH) i);
    }
    return;
}

#endif /*CHAR_WIDTH_CACHE*/

#ifndef XCOORD_CACHE
PRIVATE VOID priv_InitXCoordCache(void) {;}
PRIVATE VOID priv_ResetXCoordCache(void) {;}
PRIVATE VOID priv_InvalidateCachedLineXCoord(OPTIM LPDOC lpDoc) {;}

#else /*XCOORD_CACHE*/

/****************************************************************
	XCOORD_CACHE (on a per-line basis)
****************************************************************/	

PRIVATE VOID 	priv_BuildCachedLine	(LPDOC,LPTEXTCH,int,HMEMBLK);

#define    MAX_LINES_IN_CACHE 8

typedef struct
{
    BOOL in_use;
    int     line_num;
    HMEMBLK  hTxt;	/* for identification only  (remove?) */	
    int	    num_chars;
    int     start_pos;
    int	    attrFlush;
    HMEMBLK  hXcoord;	/* for holding x-coords */
} LREC;

#ifdef RV /*11/15*/
#define    MINIMUM_LINE_SIZE (32 * sizeof(XCOORD))
#else
#define    MINIMUM_LINE_SIZE (16 * sizeof(XCOORD))
#endif

#define    INVALID_LINE	(-1)
#define    INVALID_OFFSET  (-1)

PRIVATE LREC theCachedXCoords[MAX_LINES_IN_CACHE];

/*-------------------------------------------------*/
PRIVATE VOID priv_InitXCoordCache()
/*-------------------------------------------------*/
{
    OPTIM LREC *p;
    
    for( p = theCachedXCoords ; 
	 p < &theCachedXCoords[MAX_LINES_IN_CACHE]; 
	 p++)
    {
	p->in_use	= FALSE;
	p->line_num	= INVALID_LINE;
	p->hTxt	= NULL;	
	p->num_chars    = 0;
	p->start_pos    = INVALID_OFFSET;
	p->attrFlush		= 0;
	p->hXcoord	= mem_AllocHandle((MEMSIZE)  MINIMUM_LINE_SIZE);
    }
    priv_ResetPrevLine();
    priv_InitCharWidthCache();
}


/*-------------------------------------------------*/
PRIVATE VOID priv_ResetXCoordCache(void)
/*-------------------------------------------------*/
{
    OPTIM LREC *p;
    
    for( p = theCachedXCoords ; 
	 p < &theCachedXCoords[MAX_LINES_IN_CACHE]; 
	 p++)
    {
	p->in_use	= FALSE;
	p->line_num	= INVALID_LINE;
	p->hTxt	= NULL;	
	p->num_chars    = 0;
	p->start_pos    = INVALID_OFFSET;
	p->attrFlush		= 0;
	/* Note: p->hXcoord is left alone (to be re-used via 
	    mem_ChangeHandleSize)*/
    }
    priv_ResetPrevLine();
}

/*-------------------------------------------------*/
PRIVATE HMEMBLK priv_GetLineHandleXCoord(
    LPDOC lpDoc, 
    LPTEXTCH lpTxt,
    int line_num)
/*-------------------------------------------------*/
{
    /* NOTE: this routine assumes that hTxt is locked
     * (actually, its subordinate routine priv_BuildCachedLine()
     * is the one who makes this assumption.
     */
    OPTIM LREC *p;
    BOOL found = FALSE;
    int num_chars,start_pos;
    
    if (line_num > lpDoc->numLines) err_FatalError("GetLineHandleXCoord");
    
    start_pos = lpDoc->lineMap[line_num];     
    num_chars = (line_num < lpDoc->numLines)
		 ? lpDoc->lineMap[line_num + 1] - start_pos
	         : lpDoc->numTextChars - start_pos;

    if(num_chars < 0) err_FatalError("getlinehandle: numchars < 0!");

    for(  p = theCachedXCoords 
	; p < &theCachedXCoords[MAX_LINES_IN_CACHE]
	; p++)
    {

	if(    p->in_use
	    && p->line_num  == line_num 
	    && p->start_pos == start_pos
	    && p->hTxt     == lpDoc->hTxt)
	{
	    found = TRUE; break;
	}
    }

    if(found)
    {
	if(num_chars == p->num_chars) 
	{
	    if(p->attrFlush != lpDoc->attrFlush) 
	    {
		p->attrFlush = lpDoc->attrFlush;
		priv_BuildCachedLine(lpDoc, lpTxt,line_num, p->hXcoord);
	    }
	    return p->hXcoord;
	}
	else 
	{
	    Size size = (MEMSIZE) ((num_chars + 2) * sizeof(XCOORD));
	    
	    p->num_chars = num_chars;
	    p->attrFlush      = lpDoc->attrFlush;
	    /* don't dispose handle, just re-use it at the appropriate size */
	    if( size > MINIMUM_LINE_SIZE)
	    {
		p->hXCoord = mem_ChangeHandleSize(p->hXcoord, size);
	    }
	    priv_BuildCachedLine(lpDoc, lpTxt, line_num, p->hXcoord);
	    return p->hXcoord;
	}
     }
    /* else: line not in cache, so allocate */
    found = FALSE;

    for( p = theCachedXCoords ; p < &theCachedXCoords[MAX_LINES_IN_CACHE]; p++)
    {
	if(!p->in_use)
	{
	    p->in_use = TRUE; found = TRUE; break;
	}
    }

    if(!found)	/* a free slot was not found, re-use the first slot */
    {
	p = theCachedXCoords;
    }

    p->num_chars = num_chars;
    p->line_num  = line_num;
    p->start_pos = start_pos;
    p->hTxt     = lpDoc->hTxt;
    p->attrFlush      = lpDoc->attrFlush;

    BLOCK
    {
	Size size = (MEMSIZE) ((num_chars + 2) * sizeof(XCOORD));
	if( size > MINIMUM_LINE_SIZE)
	{
	    mem_ChangeHandleSize(p->hXcoord, size);
	}
    }
    priv_BuildCachedLine(lpDoc, lpTxt,	line_num, p->hXcoord);
    return p->hXcoord;
}

/****************************************************************/
PRIVATE VOID priv_InvalidateCachedLineXCoord(OPTIM LPDOC lpDoc)
{
    OPTIM LREC *p;
    int line_num = priv_LocateLine(lpDoc, lpDoc->selEnd);

    for(  p = theCachedXCoords 
	; p < &theCachedXCoords[MAX_LINES_IN_CACHE]
	; p++)
    {
	if( p->in_use && p->line_num==line_num) 
	{
	    p->in_use   = FALSE;
	    p->line_num	= INVALID_LINE;
	    p->hTxt	= NULL;	
	    p->num_chars= 0;
	    p->start_pos= INVALID_OFFSET;
	    p->attrFlush	= 0;
	    break;
	}
    }
}

/****************************************************************/
PRIVATE VOID  priv_BuildCachedLine(
    OPTIM LPDOC lpDoc, 
    LPTEXTCH lpTxt,
    int cur_line, 
    HMEMBLK hXcoord)
{
    int  cur_pos,end_pos,line_width;
    OPTIM LPXCOORD lpCoord;
    
    /* NOTE: this routine assumes that hTxt has been locked by caller */

    if (cur_line > lpDoc->numLines) err_FatalError("BuildCachedLine");
    
    cur_pos   = lpDoc->lineMap[cur_line];
    end_pos   = (cur_line < lpDoc->numLines)
		    ? lpDoc->lineMap[cur_line + 1]
		    : lpDoc->numTextChars;

    if(end_pos < cur_pos) err_FatalError("buildlinecache #2");
    
#ifdef MULTI_FONT
    if(lpDoc->hFormat == NULL)
#else
    BLOCK
#endif	
    {
	lpCoord = (LPXCOORD ) mem_GraspHandle(hXcoord);
	*lpCoord++ = 0; /* the zeroth element contains the line width */
	*lpCoord   = 0; /* the next element contains the zeroth char pos*/

	if(cur_pos < end_pos) 
	{
	    OPTIM LPTEXTCH lpChar;
	    LPTEXTCH end_ptr;
	    
	    priv_RenewCharWidthCache(
			lpDoc->attrTypefont,
			lpDoc->attrTypeface,
			lpDoc->attrPointsize);
	
	    for ( end_ptr = lpTxt + end_pos,
		  lpChar   = lpTxt + cur_pos
		; lpChar < end_ptr
		; lpChar++, lpCoord++)
	    {
		*(lpCoord + 1) = char_width(*lpChar) + *lpCoord;
	    }
	}
	line_width = *lpCoord; /* from above */
    }
#ifdef MULTI_FONT
    else
    {
	OPTIM LPFORMAT lpFmt;
	LPFORMAT pFormatList;
	BOOL found;
	int cur_end;
	
	lpCoord = (LPXCOORD ) mem_GraspHandle(hXcoord);
	*lpCoord++ = 0; /* the zeroth element contains the line width */
	*lpCoord   = 0; /* the next element contains the zeroth char pos*/

	mem_GraspHandle(lpDoc->hFormat);

	pFormatList = (LPFORMAT ) *(lpDoc->hFormat);

	lpFmt = ft_GetFirstFState(pFormatList, cur_pos, &found);

	if(found)   
	{
	    priv_RenewCharWidthCache(
			lpFmt->font,
			lpFmt->face,
			lpDoc->attrPointsize);
	    lpFmt = ft_GetNextFState(lpFmt, end_pos, &found);	    
	}
	else	
	{
	    priv_RenewCharWidthCache(
			    lpDoc->attrTypefont,
			    lpDoc->attrTypeface,
			    lpDoc->attrPointsize);
	    lpFmt = ft_GetNextFState(
		(LPFORMAT )(pFormatList-1),end_pos,&found);
         }	 
	 while( cur_pos < end_pos)
	 {
	     cur_end = (found && lpFmt->position != INVALID_POSITION)	    
			 ? MIN(end_pos,lpFmt->position)
			 : end_pos;
	     BLOCK
	     {
		 OPTIM LPTEXTCH lpChar;
		 LPTEXTCH end_ptr;

		 for ( end_ptr = lpTxt + cur_end,
		       lpChar   = lpTxt + cur_pos
		       ; lpChar < end_ptr
		       ; lpChar++, lpCoord++)
		 {
		     *(lpCoord + 1) = char_width(*lpChar) + *lpCoord;
		 }
	     }
	     if(found)
	     {
		 priv_RenewCharWidthCache(lpFmt->font,
				 lpFmt->face,
				 lpDoc->attrPointsize);
	     }
	     lpFmt = ft_GetNextFState((LPFORMAT)
				 (found? lpFmt : pFormatList-1), 
					    end_pos, &found);	    
	     cur_pos = cur_end;
	  }
	  line_width = *lpCoord;
	  mem_UngraspHandle(lpDoc->hFormat);
    }
#endif /*MULTI_FONT*/

    /* now that we have the accumulated offsets in the XCOORD array,
     * these need to be adjusted by a delta value that depends upon
     * whether the line is being centered, right or left justified.
     */
    BLOCK
    {
	LPXCOORD pEndOfLine;
	int dx;

	dx =  (lpDoc->attrFlush==FLUSH_LEFT ) 
			? lpDoc->galleyRect.left
	    : (lpDoc->attrFlush==FLUSH_RIGHT) 
			? lpDoc->galleyRect.right - line_width
	    : lpDoc->galleyRect.left +
	    ((((lpDoc->galleyRect.right -lpDoc->galleyRect.left) 
				- line_width)+1)>>1);

	pEndOfLine = lpCoord; /* from previous loop */

			/*lpCoord points to start of array */
	lpCoord = (LPXCOORD ) mem_GraspHandle(hXcoord); 
	*lpCoord = line_width; /* zeroth array element contains line_width*/


	lpCoord += 1;  /* zeroth char is index #1 in array */

	if(lpCoord <= pEndOfLine)    
	{
	    /* now adjust rest of elements of array */
	    for(
	       ; lpCoord <= pEndOfLine
	       ; lpCoord++)
		       *lpCoord += dx;
	}
	else
	    *lpCoord = dx;

    }
    mem_UngraspHandle(hXcoord);
}
#endif    

/****************************************************************/
PRIVATE int priv_CalcLinebreak(
    OPTIM LPTEXTCH 	lpChar,
    OPTIM LPTEXTCH 	end_of_block,
    int			spec_measure)
{
    LPTEXTCH 	saved_position;
    LPTEXTCH 	start_position;
    BOOL	space_encountered;
    int		accum_measure;

    saved_position    = start_position = lpChar;
    space_encountered = FALSE; 
    accum_measure     = 0;


    /*priv_RenewCharWidthCache();  <<< needs to be added here? <<<*/

    for (
	; lpChar < end_of_block
        ; lpChar++)
    {
		    
        accum_measure += char_width(*lpChar);

        if (accum_measure >= spec_measure)
        {
	    /* break line either at current char or at one
	     * position AFTER the position of the last space
	     */
	    if(space_encountered)
	    {
		lpChar = saved_position + 1;
	    }
	    break;
        }
        else if(*lpChar==NEWLINE_CH)
        {
	    lpChar++; /* break line AFTER the newline */
	    break;
        }
        else if(*lpChar==SPACE_CH)		
        {
	    saved_position    = lpChar;
	    space_encountered = TRUE;
	    /*continue*/
        }
	/* else: must be regular character, so continue */
	
    }  /* for-loop */

	/* make sure there is at least one character per line */
    return((lpChar > start_position)
		? (int) (lpChar - start_position) 
		: 1 );
}

/****************************************************************
    ED_RFMT.C -- incremental reformatting
****************************************************************/

PRIVATE int prev_cur_line;

/*-------------------------------------------------*/
PRIVATE VOID priv_ResetPrevLine(void)
/*-------------------------------------------------*/
{
    prev_cur_line = INVALID_LINE;
#ifdef OUTPUT_BITMAP_CACHE
    is_bitmap_valid = FALSE;
#endif
}

/*-------------------------------------------------*/
PRIVATE VOID priv_IncrementalReformat(
    LPDOC	lpDoc,
    int		prev_bottom)
/*-------------------------------------------------*/
{
    /* incremental reformat =AND= redraw of affected area. */
    int		cur_line;
    int		y;


    priv_ShowStructures(lpDoc);

    /* First, find the text position to start reformatting.
     * In parallel, find the starting y-location
     */
    for ( cur_line = 0, y = lpDoc->galleyRect.top
	; cur_line <= lpDoc->numLines
	; cur_line++, y += lpDoc->lineHeight
	)
    {
	/* we want to reformat when either:
	 * 	- the current linestart marker has been declared invalid, OR
	 *	- the current text position (selEnd) has been reached
	 */
	if (lpDoc->lineMap[cur_line]==LS_INVALID_POSITION) 
	{
	    if (cur_line > 0) { cur_line--; y -= lpDoc->lineHeight; }
	    break;
	}
	else if ( lpDoc->lineMap[cur_line + 1] > lpDoc->selEnd) 
	    break;
    }
    if (cur_line==0)    lpDoc->lineMap[cur_line] = 0;

    mem_GraspHandle(lpDoc->hTxt); /* lock handle for the following function */

    priv_FormatUntilSynch(hDoc,cur_line,y,prev_bottom);

    mem_UngraspHandle(lpDoc->hTxt);

}

/*-------------------------------------------------*/
PRIVATE VOID priv_DeleteOneChar(
    OPTIM LPDOC lpDoc,
    int		prev_bottom)
/*-------------------------------------------------*/
{
    HMEMBLK	hTxt;

    /* This routine (or its subordinates) may unlock hDoc and 
     * mem_ChangeHandleSize it, so any calling routines should renew their
     * pointers to lpDoc)
     */

#ifdef MULTI_FONT
    if(lpDoc->hFormat)
	ft_DeleteFromFormat(lpDoc,lpDoc->selEnd - 1, lpDoc->selEnd);
#endif

    hTxt = lpDoc->hTxt;
    lpTxt = (LPTEXTCH) mem_GraspHandle(hTxt);

    /* shift the bytes over to delete the character */
    BLOCK
    {
	int		cur_pos = lpDoc->selEnd;
	OPTIM LPTEXTCH 	src = lpTxt + cur_pos;
	OPTIM LPTEXTCH 	dst = lpTxt + cur_pos - 1;
	OPTIM int	cnt = lpDoc->numTextChars - cur_pos;
    
	while(cnt--) *dst++ = *src++;

	/* mem_ChangeHandleSize() is called at end of this function,
	 * after unlocking hTxt
	 */
    }

    /* both insert_one_ch() and delepriv_one_ch() call priv_ReformatForOne */

    /* hTxt must be locked for the following routine */
    priv_ReformatForOne(hDoc,hTxt,-1,prev_bottom); /*delta length is -1 */

    mem_UngraspHandle(hTxt);
    
    lpDoc->hTxt = mem_ChangeHandleSize(hTxt, 
	(MEMSIZE) (MAX(MIN_HTXT_SIZE,(lpDoc->numTextChars))));
    return;
}

/*-------------------------------------------------*/
PRIVATE VOID priv_InsertOneChar(
    LPDOC	lpDoc,
    TEXTCH	ch,
    int		prev_bottom)
/*-------------------------------------------------*/
{
    HMEMBLK	hXcoordMap;
    HMEMBLK	hTxt;
    LPTEXTCH    lpTxt;
    
    hXcoordMap = lpDoc->hXcoordMap;
    hTxt = lpDoc->hTxt;


#ifdef MULTI_FONT
    if(lpDoc->hFormat)
	ft_AddToFormat(lpDoc,lpDoc->selEnd,1);
#endif

    /* for efficiency, we could batch up these calls by making the 
     * text block grow in discrete increments.
     */
    hTxt = lpDoc->hTxt = mem_ChangeHandleSize(hTxt, 
	    (MEMSIZE) (MAX(MIN_HTXT_SIZE,(lpDoc->numTextChars+1))));
    
    lpTxt = (LPTEXTCH) mem_GraspHandle(hTxt);

    /* shift the bytes over to insert the character */
    BLOCK
    {
	int		cur_pos = lpDoc->selEnd;
	OPTIM LPTEXTCH 	src = lpTxt + lpDoc->numTextChars - 1;
	OPTIM LPTEXTCH 	dst = lpTxt + lpDoc->numTextChars    ;
	OPTIM int	cnt = lpDoc->numTextChars - cur_pos;
    
	while(cnt--) *dst-- = *src--;

	*dst = ch;  /* put the char in the hole */
    }

    /* both insert_one_ch() and delepriv_one_ch() call priv_ReformatForOne */
    
    priv_ReformatForOne(hDoc,hTxt,1,prev_bottom); /*delta length is 1 */

    mem_UngraspHandle(hTxt);
    
    return;
}
/*-------------------------------------------------*/
PRIVATE int priv_UpdateLineStarts(
    OPTIM LPDOC	lpDoc,
    int		delta_length)
/*-------------------------------------------------*/
{
    int cur_pos,cur_line;

    /* update the value of numTextChars to reflect the delta value
     * Update other variables: selStart,selEnd,cur_pos.
     */

    lpDoc->numTextChars += delta_length;
    lpDoc->selEnd       += delta_length;
    lpDoc->selStart      = lpDoc->selEnd;
    cur_pos              = lpDoc->selEnd;

    /* find out the values of:
     *	-- the variable cur_line (i.e., index into lineMap array)
     *  -- the corresponding y-position of that line
     *
     * These values could be cached from invocation to invocation
     * to save on calculation in the case of very long text blocks.
     */
    if(cur_pos >= lpDoc->numTextChars) /* on last char of last line */
    {
	cur_line = MAX(lpDoc->numLines - 1,0);
	if(lpDoc->numLines==0)
	{
	    lpDoc->numLines = 1;
	    lpDoc->lineMap[1] = lpDoc->numTextChars;
	}
	/* add delta length to subsequent linestarts */
	lpDoc->lineMap[lpDoc->numLines] += delta_length;
    }
    else
    {
	OPTIM int i;
	
	for (   cur_line = -1, i = 0
	      ; i < lpDoc->numLines 
	      ; i++
	    )
	{
	    if (lpDoc->lineMap[i + 1] >= cur_pos) 
	    {
		cur_line = i; break;
	    }
	}
	if (cur_line== -1) err_FatalError("updlinestrt: curline not found!");

	/* add one to the subsequent line starts */
	for (  i += 1		/* skip the current line */
	     ; i <= lpDoc->numLines	/* include line after the last line*/
	     ; i++
	    )
	{
	    lpDoc->lineMap[i] += delta_length;
	}
    }
    return cur_line;
}


/*-------------------------------------------------*/
PRIVATE VOID	priv_ReformatForOne(
    OPTIM LPDOC lpDoc,
    HMEMBLK	hTxt,  /* Note: this handle has been locked by caller */
    int		delta_length,
    int		prev_bottom)
/*-------------------------------------------------*/
{
    int 	num_phys_lines;
    int		snapped_view_bott;
    int		y;
    int		spec_measure;
    int		cur_line;
    LPTEXTCH 	end_of_block;
    LPTEXTCH 	lpChar;
    
    spec_measure = lpDoc->galleyRect.right - lpDoc->galleyRect.left;
    if(spec_measure <= 0) err_FatalError("refmt1: empty galleyRect!");

    cur_line = priv_UpdateLineStarts(lpDoc, delta_length);

    y = lpDoc->galleyRect.top + (cur_line * lpDoc->lineHeight);
	
    /* line starts now correspond to text positions,
     * but formatting (line-breaks) are possibly invalid.
     * Set up some formatting information.
     */
    num_phys_lines = lpDoc->numPhysLines;
    snapped_view_bott  = lpDoc->viewRect.bottom
	- ((lpDoc->viewRect.bottom - lpDoc->galleyRect.top) 
			    % lpDoc->lineHeight)
	- lpDoc->lineHeight;	
    end_of_block   = lpTxt + lpDoc->numTextChars;	
    
	
    /* if there is a previous line, re-calculate the line break for that
     * one first.
     */
    if (cur_line > 0)
    {
	LPTEXTCH cur_break;
	
	lpChar     = lpTxt + lpDoc->lineMap[cur_line-1];
	cur_break  = lpTxt + lpDoc->lineMap[cur_line  ];
	
	 /* figure out where the line will break */
	 lpChar += priv_CalcLinebreak(lpChar, end_of_block, spec_measure);

	 /* if the line breaks at a different place from before
	  * record the new line break and, if necessary, draw the line.
	  */
	 if(lpChar != cur_break)
	 {
	     lpDoc->lineMap[cur_line] = (int)(lpChar - lpTxt);

	    if(    y <= snapped_view_bott 
		&& y - lpDoc->lineHeight >= lpDoc->viewRect.top)
	     {
		 prev_cur_line = INVALID_LINE;
#ifdef OUTPUT_BITMAP_CACHE
		 is_bitmap_valid = FALSE;
#endif
		/* don't need to check if this node is active before
		 * clearing the background,because this routine is only
		 * called by priv_OnCharMsg which only occurs if text is active.
		 */
		 gp_ClearRect( lpDoc->viewRect.left, y - lpDoc->lineHeight,
			     lpDoc->viewRect.right,lpDoc->lineHeight);

		 gp_DrawLineOfText(lpDoc, hTxt, 
				 cur_line - 1, y - lpDoc->lineHeight,FALSE);

		 /*clear_line_cache(); don't save this line */

		 if(lpChar >= end_of_block)
		 {

		     gp_ClearRect( lpDoc->viewRect.left,y,
				 lpDoc->viewRect.right,lpDoc->lineHeight);

		     lpDoc->numLines -= 1;
		 }
	     }
	 }
    }
    priv_FormatUntilSynch(hDoc,cur_line,y, prev_bottom);
    return;
}

/*-------------------------------------------------*/
PRIVATE VOID	priv_FormatUntilSynch(
    OPTIM LPDOC	lpDoc,
    int		cur_line,
    int		y,
    int		prev_bottom)
/*-------------------------------------------------*/
{
    int		num_phys_lines;
    int		spec_measure;
    HMEMBLK	hTxt;
    BOOL	synch_found;
    LPTEXTCH 	lpChar;
    LPTEXTCH 	end_of_block;
    int		snapped_view_bott;
    BOOL	do_draw_line = TRUE;
    int		starting_line = cur_line;
    
    hTxt = lpDoc->hTxt; /* has been locked by calling function*/

    snapped_view_bott  = 
	lpDoc->viewRect.bottom
	    - ((lpDoc->viewRect.bottom - lpDoc->galleyRect.top) 
					    % lpDoc->lineHeight)
	    - lpDoc->lineHeight;	

    num_phys_lines = lpDoc->numPhysLines;
    spec_measure   = lpDoc->galleyRect.right - lpDoc->galleyRect.left;

    if(    (lpDoc->viewRect.bottom < lpDoc->galleyRect.top)
	|| (spec_measure <= 0) )
	err_FatalError("fmtsynch");
    
    for (  lpChar        = lpTxt + lpDoc->lineMap[cur_line],
           end_of_block = lpTxt + lpDoc->numTextChars,
	   synch_found = FALSE
	 ; lpChar < end_of_block && !synch_found
	 ; cur_line++, y += lpDoc->lineHeight
	) 
    {
	 /* figure out where the line will break */
	 lpChar += priv_CalcLinebreak(lpChar, end_of_block, spec_measure);

	 if( cur_line + 1 >= num_phys_lines)
	 {
	     mem_UngraspHandle( hLineMap);

	     lpDoc->hLineMap = hLineMap = mem_ChangeHandleSize( 
		 hLineMap,
		 (MEMSIZE) (mem_GetHandleSize( hLineMap ) + 
				 (LINES_QUANTUM * sizeof(TEXTPOS))));

	     /*regenerate lpTxt ptr, which is possibly invalid now*/
	     lpTxt = (LPTEXTCH) mem_GraspHandle( hDoc);

	     num_phys_lines += LINES_QUANTUM;
	 }

	 /* test if line breaks match. If so, we are in synch, so quit. */
	 if(lpDoc->lineMap[cur_line+1] ==(TEXTPOS)(lpChar - lpTxt);
	 {
	     synch_found = TRUE;  /* will break at top of loop */
	 }
	 else
	 {
	     /* else record line break and move to next line */
	     lpDoc->lineMap[cur_line+1] = (TEXTPOS)(lpChar - lpTxt);

	 }
	 if( y <= snapped_view_bott)
	 {
	     /* draw a line if we are within snapped view rect bottom,
	      * and if we satisfy the caching requirements.
	      */
	     if(cur_line==starting_line && cur_line==prev_cur_line)
	     {
		 if(kb_IsThereKeystroke(lpDoc))  do_draw_line = FALSE;
		 else
		 {
		     /* don't need to check if this node is active before
		      * clearing the background,because this routine is only
		      * called by priv_OnCharMsg which only occurs if text 
		      * is active.
		      */
		     gp_ClearRect( 
			 lpDoc->viewRect.left,
			 y,
			 lpDoc->viewRect.right,
			 lpDoc->lineHeight);
#ifndef OUTPUT_BITMAP_CACHE
		     gp_DrawLineOfText(lpDoc,hTxt, cur_line, y, FALSE);
#else
		     gp_DrawLineOfText(lpDoc,hTxt, cur_line, y, TRUE);
#endif		 
		 }
	     }
	     else if(do_draw_line)
	     {
		 gp_ClearRect( lpDoc->viewRect.left,y,
			     lpDoc->viewRect.right,lpDoc->lineHeight);
		 gp_DrawLineOfText(lpDoc,hTxt, cur_line, y, FALSE);
	     }
	 }
    }

    if( !synch_found)	
    {
	lpDoc->numLines = cur_line; 
	prev_cur_line = INVALID_LINE;
	if( y - lpDoc->lineHeight <= prev_bottom)
	{
	    gp_ClearRect(
	       lpDoc->viewRect.left,
	       y,
	       lpDoc->viewRect.right,
	       (prev_bottom + lpDoc->lineHeight - y));
	}
    }
    else if(cur_line-1 == starting_line) /*synch found and curln==startln*/
    {
	prev_cur_line = cur_line-1;
    }
    lpDoc->numPhysLines  = num_phys_lines;
    lpDoc->lineMap[lpDoc->numLines] = lpDoc->numTextChars;    
    
    return;    
}



/*-------------------------------------------------*/
PRIVATE VOID	priv_MarkLineStarts(
    LINEOP 	op,
    OPTIM LPDOC	lpDoc,
    int		start,
    int 	end)
/*-------------------------------------------------*/
{
    /* NOTE: this assumes lineMap are valid
     * and that the DOCSTRUC pointed to by hDoc has been locked
     */
    int	cur_line, start_line, end_line, segment_length;

    segment_length = end - start;
    start_line = priv_LocateLine(lpDoc,start);

    if( op==LS_DELETE_SEGMENT )
    {
	end_line   = priv_LocateLine(lpDoc, end);
    
	for ( cur_line = start_line
	    ; cur_line <= end_line
	    ; cur_line++
	    )
	{
	   lpDoc->lineMap[cur_line] = LS_INVALID_POSITION;
	}

	for ( /* use cur_line value from above */
	    ; cur_line <= lpDoc->numLines
	    ; cur_line++
	    )
	{
	    lpDoc->lineMap[cur_line] -= segment_length;
	}
    }
    else if( op==LS_INSERT_SEGMENT )
    {
	lpDoc->lineMap[start_line] = LS_INVALID_POSITION;

	for ( cur_line  = start_line+1  /* the line AFTER start_line */
	    ; cur_line <= lpDoc->numLines /* include the line aftr the last */
	    ; cur_line++
	    )
	{
	    lpDoc->lineMap[cur_line] += segment_length;
	}

    }
}
/*---------------------------------------------------------*/
PUBLIC VOID ed_OnIdleMsg(HDOC hDoc)
/*---------------------------------------------------------*/
{
    /* for now, not much to do here*/;
}


/****************************************************************
	    OUTPUT_BITMAP_CACHE
******************************************************************/
#ifdef OUTPUT_BITMAP_CACHE

PRIVATE BOOL is_bitmap_valid;

PRIVATE HDC     hScreenDC;
PRIVATE HDC     hSavedLineDC;
PRIVATE HBITMAP theLineImage;

/*---------------------------------------------------------*/
PRIVATE VOID 	priv_InitBitmapCache(void)
/*---------------------------------------------------------*/
{
  hScreenDC = CreateDC((LPSTR)"DISPLAY",(LPSTR)0,(LPSTR)0,(LPSTR)0);
  if(hScreenDC==NULL) err_FatalError("text: CreateDC failed!");
  hSavedLineDC = CreateCompatibleDC(hScreenDC);
  if(hSavedLineDC==NULL) err_FatalError("text: CreateCompatibleDC failed!");
}


/*---------------------------------------------------------*/
PRIVATE VOID  priv_ResetBitmapCache(void)
/*---------------------------------------------------------*/
{
    is_bitmap_valid = FALSE;

}

/*---------------------------------------------------------*/
PRIVATE VOID  priv_FiniBitmapCache(void)
/*---------------------------------------------------------*/
{
    /*for now, not much to do here*/;
}

/*---------------------------------------------------------*/
PRIVATE VOID  priv_CacheOneBitmap(x,y,width,depth)
    int x,y,width,depth;
/*---------------------------------------------------------*/
{
    theLineImage = win_CreateBitmap( 
		      width,		/* width */	
		      depth,		/* depth */
		      1, 		/* nPlanes */
		      1,		/* bit count */
		      (LPSTR) NULL);	/* initialize bits or not (if null) */

    win_DeleteObject( win_SelectObject(hSavedLineDC, theLineImage));
    win_StretchBlt(
	   hSavedLineDC , 0,  0, width, depth,
	   hScreenDC,     win_pt.x,  win_pt.y, width, depth,
	   SRCCOPY
	   );  
}

/*---------------------------------------------------------*/
PRIVATE VOID  priv_DrawOneBitmap(
    int x,
    int y,
    int width,
    int depth)
/*---------------------------------------------------------*/
{
    POINT pt;

    transform_coordinates(x,y,&pt);
    transform_coordinates(width,depth,&pt2);    

    if(!theLineImage) err_FatalError("draw bitmap");
    win_StretchBlt(
	   hScreenDC ,    pt.x,  pt.y, pt2.x, pt2.y,
	   hSavedLineDC,  0,  0, width, depth,
	   SRCCOPY
	   );  
}
#endif /*OUTPUT_BITMAP_CACHE*/



